from DIRAC import gLogger
from DIRAC.Resources.Storage.GFAL2_XROOTStorage import GFAL2_XROOTStorage
from DIRAC.Resources.Storage.GFAL2_SRM2Storage import GFAL2_SRM2Storage

sLog = gLogger.getSubLogger(__name__)


class CTAStorage(GFAL2_XROOTStorage):
    """Plugin to interact with CERN CTA.

    It basically is XROOT with added tape capabilities.
    Since CTA supports ONLY xroot, do not forget to add
    xroot in your `Operations/<Setup>/DataManagement/RegistrationProtocols` list

    Configuration example::

        StorageElements
        {
          CTA-PPS
          {
            BackendType = Cta
            AccessProtocols = root
            WriteProtocols = root
            # This is very important if you have to stage with this protocol,  but might transfer
            # using a different protocol, like https
            StageProtocols = root
            SEType = T1D0
            SpaceReservation = LHCb-Tape
            OccupancyLFN = /eos/ctalhcbpps/proc/accounting
            OccupancyPlugin = WLCGAccountingJson
            # Config for this plugin is below
            ###################################
            CTA
            {
              Host = eosctalhcbpps.cern.ch
              Protocol = root
              Path = /eos/ctalhcbpps/archivetest/
              Access = remote
            }
            ###################################
            GFAL2_HTTPS
            {
              Host = eosctalhcbpps.cern.ch
              Protocol = https
              Path = /eos/ctalhcbpps/archivetest/
              Access = remote
            }
          }
        }
    """

    # Copy from SRM the method that updates the metadata
    # info with the tape specific information (Cached, Migrated, etc)
    # Note: `meth = Class.meth` does not work because it would assign to B
    # an unbound method of A.
    _updateMetadataDict = GFAL2_SRM2Storage.__dict__["_updateMetadataDict"]

    def __init__(self, storageName, parameters):
        """c'tor

        :param self: self reference
        :param str storageName: SE name
        :param dict parameters: passed to parent's class
        """
        # # init base class
        super(CTAStorage, self).__init__(storageName, parameters)

        self.log = sLog.getSubLogger(storageName)

        self.pluginName = "CTA"

        # We need user.status for Tape metadata
        self._defaultExtendedAttributes = ["user.status"]



########################################################################
# File :    CPUNormalization.py
# Author :  Ricardo Graciani
########################################################################

""" DIRAC Workload Management System Client module that encapsulates all the
    methods necessary to handle CPU normalization
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from six.moves.urllib.request import urlopen

import DIRAC
from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import TimeLeft
from DIRAC.WorkloadManagementSystem.Client.DIRACbenchmark import singleDiracBenchmark

__RCSID__ = "$Id$"

# TODO: This should come from some place in the configuration
NORMALIZATIONCONSTANT = 60.0 / 250.0  # from minutes to seconds and from SI00 to HS06 (ie min * SI00 -> sec * HS06 )

UNITS = {"HS06": 1.0, "SI00": 1.0 / 250.0}

# TODO: This is still fetching directly from MJF rather than going through
# the MJF module and the values it saves in the local DIRAC configuration


def __getFeatures(envVariable, items):
    """Extract features"""
    features = {}
    featuresDir = os.environ.get(envVariable)
    if featuresDir is None:
        return features
    for item in items:
        fname = os.path.join(featuresDir, item)
        try:
            # Only keep features that do exist
            features[item] = urlopen(fname).read()
        except IOError:
            pass
    return features


def getMachineFeatures():
    """This uses the _old_ MJF information"""
    return __getFeatures("MACHINEFEATURES", ("hs06", "jobslots", "log_cores", "phys_cores"))


# TODO: log_cores and phys_cores are deprecated and from old MJF specificationa and not collected
# by the MJF module!


def getJobFeatures():
    """This uses the _new_ MJF information"""
    return __getFeatures("JOBFEATURES", ("hs06_job", "allocated_cpu"))


def getPowerFromMJF():
    """Extracts the machine power from either JOBFEATURES or MACHINEFEATURES"""
    try:
        features = getJobFeatures()
        hs06Job = features.get("hs06_job")
        # If the information is there and non zero, return, otherwise go to machine features
        if hs06Job:
            return round(float(hs06Job), 2)
        features = getMachineFeatures()
        totalPower = float(features.get("hs06", 0))
        logCores = float(features.get("log_cores", 0))
        physCores = float(features.get("phys_cores", 0))
        jobSlots = float(features.get("jobslots", 0))
        denom = min(max(logCores, physCores), jobSlots) if (logCores or physCores) and jobSlots else None
        if totalPower and denom:
            return round(totalPower / denom, 2)
        return None
    except ValueError as e:
        gLogger.exception("Exception getting MJF information", lException=e)
        return None


def queueNormalizedCPU(ceUniqueID):
    """Report Normalized CPU length of queue"""
    result = getQueueInfo(ceUniqueID)
    if not result["OK"]:
        return result

    ceInfoDict = result["Value"]
    siteCSSEction = ceInfoDict["SiteCSSEction"]
    queueCSSection = ceInfoDict["QueueCSSection"]

    benchmarkSI00 = __getQueueNormalization(queueCSSection, siteCSSEction)
    maxCPUTime = __getMaxCPUTime(queueCSSection)

    if maxCPUTime and benchmarkSI00:
        normCPUTime = NORMALIZATIONCONSTANT * maxCPUTime * benchmarkSI00
    else:
        if not benchmarkSI00:
            subClusterUniqueID = ceInfoDict["SubClusterUniqueID"]
            return S_ERROR("benchmarkSI00 info not available for %s" % subClusterUniqueID)
        if not maxCPUTime:
            return S_ERROR("maxCPUTime info not available")

    return S_OK(normCPUTime)


def getQueueNormalization(ceUniqueID):
    """Report Normalization Factor applied by Site to the given Queue"""
    result = getQueueInfo(ceUniqueID)
    if not result["OK"]:
        return result

    ceInfoDict = result["Value"]
    siteCSSEction = ceInfoDict["SiteCSSEction"]
    queueCSSection = ceInfoDict["QueueCSSection"]
    subClusterUniqueID = ceInfoDict["SubClusterUniqueID"]

    benchmarkSI00 = __getQueueNormalization(queueCSSection, siteCSSEction)

    if benchmarkSI00:
        return S_OK(benchmarkSI00)
    return S_ERROR("benchmarkSI00 info not available for %s" % subClusterUniqueID)
    # errorList.append( ( subClusterUniqueID , 'benchmarkSI00 info not available' ) )
    # exitCode = 3


def __getQueueNormalization(queueCSSection, siteCSSEction):
    """Query the CS and return the Normalization"""
    benchmarkSI00Option = "%s/%s" % (queueCSSection, "SI00")
    benchmarkSI00 = gConfig.getValue(benchmarkSI00Option, 0.0)
    if not benchmarkSI00:
        benchmarkSI00Option = "%s/%s" % (siteCSSEction, "SI00")
        benchmarkSI00 = gConfig.getValue(benchmarkSI00Option, 0.0)

    return benchmarkSI00


def __getMaxCPUTime(queueCSSection):
    """Query the CS and return the maxCPUTime"""
    maxCPUTimeOption = "%s/%s" % (queueCSSection, "maxCPUTime")
    maxCPUTime = gConfig.getValue(maxCPUTimeOption, 0.0)
    # For some sites there are crazy values in the CS
    maxCPUTime = max(maxCPUTime, 0)
    maxCPUTime = min(maxCPUTime, 86400 * 12.5)

    return maxCPUTime


def getCPUNormalization(reference="HS06", iterations=1):
    """Get Normalized Power of the current CPU in [reference] units"""
    if reference not in UNITS:
        return S_ERROR("Unknown Normalization unit %s" % str(reference))
    try:
        max(min(int(iterations), 10), 1)
    except (TypeError, ValueError) as x:
        return S_ERROR(x)

    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

    corr = Operations().getValue("JobScheduling/CPUNormalizationCorrection", 1.0)

    result = singleDiracBenchmark(iterations)

    if result is None:
        return S_ERROR("Cannot get benchmark measurements")

    return S_OK({"CPU": result["CPU"], "WALL": result["WALL"], "NORM": result["NORM"] / corr, "UNIT": reference})


def getCPUTime(cpuNormalizationFactor):
    """Trying to get CPUTime left for execution (in seconds).

    It will first look to get the work left looking for batch system information useing the TimeLeft utility.
    If it succeeds, it will convert it in real second, and return it.

    If it fails, it tries to get it from the static info found in CS.
    If it fails, it returns the default, which is a large 9999999, that we may consider as "Infinite".

    This is a generic method, independent from the middleware of the resource if TimeLeft doesn't return a value

    args:
      cpuNormalizationFactor (float): the CPU power of the current Worker Node.
      If not passed in, it's get from the local configuration

    returns:
      cpuTimeLeft (int): the CPU time left, in seconds
    """
    cpuTimeLeft = 0.0
    cpuWorkLeft = gConfig.getValue("/LocalSite/CPUTimeLeft", 0)

    if not cpuWorkLeft:
        # Try and get the information from the CPU left utility
        result = TimeLeft().getTimeLeft()
        if result["OK"]:
            cpuWorkLeft = result["Value"]

    if cpuWorkLeft > 0:
        # This is in HS06sseconds
        # We need to convert in real seconds
        if not cpuNormalizationFactor:  # if cpuNormalizationFactor passed in is 0, try get it from the local cfg
            cpuNormalizationFactor = gConfig.getValue("/LocalSite/CPUNormalizationFactor", 0.0)
        if cpuNormalizationFactor:
            cpuTimeLeft = cpuWorkLeft / cpuNormalizationFactor

    if not cpuTimeLeft:
        # now we know that we have to find the CPUTimeLeft by looking in the CS
        # this is not granted to be correct as the CS units may not be real seconds
        gridCE = gConfig.getValue("/LocalSite/GridCE")
        ceQueue = gConfig.getValue("/LocalSite/CEQueue")
        if not ceQueue:
            # we have to look for a ceQueue in the CS
            # A bit hacky. We should better profit from something generic
            gLogger.warn("No CEQueue in local configuration, looking to find one in CS")
            siteName = DIRAC.siteName()
            queueSection = "/Resources/Sites/%s/%s/CEs/%s/Queues" % (siteName.split(".")[0], siteName, gridCE)
            res = gConfig.getSections(queueSection)
            if not res["OK"]:
                raise RuntimeError(res["Message"])
            queues = res["Value"]
            cpuTimes = [gConfig.getValue(queueSection + "/" + queue + "/maxCPUTime", 9999999.0) for queue in queues]
            # These are (real, wall clock) minutes - damn BDII!
            cpuTimeLeft = min(cpuTimes) * 60
        else:
            queueInfo = getQueueInfo("%s/%s" % (gridCE, ceQueue))
            cpuTimeLeft = 9999999.0
            if not queueInfo["OK"] or not queueInfo["Value"]:
                gLogger.warn("Can't find a CE/queue, defaulting CPUTime to %d" % cpuTimeLeft)
            else:
                queueCSSection = queueInfo["Value"]["QueueCSSection"]
                # These are (real, wall clock) minutes - damn BDII!
                cpuTimeInMinutes = gConfig.getValue("%s/maxCPUTime" % queueCSSection, 0.0)
                if cpuTimeInMinutes:
                    cpuTimeLeft = cpuTimeInMinutes * 60.0
                    gLogger.info("CPUTime for %s: %f" % (queueCSSection, cpuTimeLeft))
                else:
                    gLogger.warn(
                        "Can't find maxCPUTime for %s, defaulting CPUTime to %f" % (queueCSSection, cpuTimeLeft)
                    )

    return int(cpuTimeLeft)


def getQueueInfo(ceUniqueID, diracSiteName=""):
    """
    Extract information from full CE Name including associate DIRAC Site
    """
    try:
        subClusterUniqueID = ceUniqueID.split("/")[0].split(":")[0]
        queueID = ceUniqueID.split("/")[1]
    except IndexError:
        return S_ERROR("Wrong full queue Name")

    if not diracSiteName:
        gLogger.debug("SiteName not given, looking in /LocaSite/Site")
        diracSiteName = gConfig.getValue("/LocalSite/Site", "")

        if not diracSiteName:
            gLogger.debug("Can't find LocalSite name, looking in CS")
            result = getCESiteMapping(subClusterUniqueID)
            if not result["OK"]:
                return result
            diracSiteName = result["Value"][subClusterUniqueID]

            if not diracSiteName:
                gLogger.error("Can not find corresponding Site in CS")
                return S_ERROR("Can not find corresponding Site in CS")

    gridType = diracSiteName.split(".")[0]

    siteCSSEction = "/Resources/Sites/%s/%s/CEs/%s" % (gridType, diracSiteName, subClusterUniqueID)
    queueCSSection = "%s/Queues/%s" % (siteCSSEction, queueID)

    resultDict = {
        "SubClusterUniqueID": subClusterUniqueID,
        "QueueID": queueID,
        "SiteName": diracSiteName,
        "Grid": gridType,
        "SiteCSSEction": siteCSSEction,
        "QueueCSSection": queueCSSection,
    }

    return S_OK(resultDict)



""" tests for the JobDB module """

# pylint: disable=protected-access, missing-docstring

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import unittest

from mock import MagicMock, patch

from DIRAC import S_OK

MODULE_NAME = "DIRAC.WorkloadManagementSystem.DB.JobDB"


class JobDBTest(unittest.TestCase):
    def setUp(self):
        def mockInit(self):
            self.log = MagicMock()
            self.logger = MagicMock()
            self._connected = True

        from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB

        with patch(MODULE_NAME + ".JobDB.__init__", new=mockInit):
            self.jobDB = JobDB()
        self.jobDB._query = MagicMock(name="Query")
        self.jobDB._escapeString = MagicMock(return_value=S_OK())

    def tearDown(self):
        pass

    def test_getInputData(self):
        self.jobDB._query.return_value = S_OK((("/vo/user/lfn1",), ("LFN:/vo/user/lfn2",)))
        result = self.jobDB.getInputData(1234)
        print(result)
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], ["/vo/user/lfn1", "/vo/user/lfn2"])



#!/usr/bin/env python
"""
Get informations for a given production

Example:
  $ dirac-prod-get 381
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("prodID: Production ID")
    _, args = Script.parseCommandLine()

    from DIRAC.Core.Utilities.PrettyPrint import printTable
    from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient

    prodClient = ProductionClient()

    # get arguments
    prodID = args[0]
    res = prodClient.getProduction(prodID)

    fields = ["ProductionName", "Status", "ProductionID", "CreationDate", "LastUpdate", "AuthorDN", "AuthorGroup"]
    records = []

    if res["OK"]:
        prodList = res["Value"]
        if not isinstance(res["Value"], list):
            prodList = [res["Value"]]
        for prod in prodList:
            records.append(
                [
                    str(prod["ProductionName"]),
                    str(prod["Status"]),
                    str(prod["ProductionID"]),
                    str(prod["CreationDate"]),
                    str(prod["LastUpdate"]),
                    str(prod["AuthorDN"]),
                    str(prod["AuthorGroup"]),
                ]
            )
    else:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    printTable(fields, records)

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" POOL XML Catalog Class
    This class handles simple XML-based File Catalog following the
    POOL project schema. It presents a DIRAC generic File Catalog interface
    although not complete and with several extensions
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
import xml.dom.minidom
from DIRAC import S_OK, S_ERROR


class PoolFile(object):
    """
    A Pool XML File Catalog entry

    @author A.Tsaregorodtsev
    """

    def __init__(self, dom=None):

        self.guid = ""
        self.pfns = []
        self.lfns = []

        if dom:
            self.guid = dom.getAttribute("ID")
            physs = dom.getElementsByTagName("physical")
            for p in physs:
                pfns = p.getElementsByTagName("pfn")
                meta = p.getElementsByTagName("metadata")
                for pfn in pfns:
                    ftype = pfn.getAttribute("filetype")
                    name = pfn.getAttribute("name")
                    # Get the SE name if any
                    se = pfn.getAttribute("se")
                    se = se if se else "Unknown"

                    self.pfns.append((name, ftype, se))
            logics = dom.getElementsByTagName("logical")
            for l in logics:
                # do not know yet the Pool lfn xml schema
                lfns = l.getElementsByTagName("lfn")
                for lfn in lfns:
                    name = lfn.getAttribute("name")
                self.lfns.append(name)

    def dump(self):
        """Dumps the contents to the standard output"""

        print("\nPool Catalog file entry:")
        print("   guid:", self.guid)
        if len(self.lfns) > 0:
            print("   lfns:")
            for l in self.lfns:
                print("     ", l)
        if len(self.pfns) > 0:
            print("   pfns:")
            for p in self.pfns:
                print("     ", p[0], "type:", p[1], "SE:", p[2])

    def getPfns(self):
        """Retrieves all the PFNs"""
        result = []
        for p in self.pfns:
            result.append((p[0], p[2]))

        return result

    def getLfns(self):
        """Retrieves all the LFNs"""
        result = []
        for l in self.lfns:
            result.append(l)

        return result

    def addLfn(self, lfn):
        """Adds one LFN"""
        self.lfns.append(lfn)

    def addPfn(self, pfn, pfntype=None, se=None):
        """Adds one PFN"""
        sename = "Unknown"
        if se:
            sename = se

        if pfntype:
            self.pfns.append((pfn, pfntype, sename))
        else:
            self.pfns.append((pfn, "ROOT_All", sename))

    def toXML(self, metadata):
        """Output the contents as an XML string"""

        doc = xml.dom.minidom.Document()

        fileElt = doc.createElement("File")
        fileElt.setAttribute("ID", self.guid)
        if self.pfns:
            physicalElt = doc.createElement("physical")
            fileElt.appendChild(physicalElt)
            for p in self.pfns:
                pfnElt = doc.createElement("pfn")
                physicalElt.appendChild(pfnElt)

                # To properly escape <>& in POOL XML slice.
                fixedp = p[0].replace("&", "&amp;")
                fixedp = fixedp.replace("&&amp;amp;", "&amp;")
                fixedp = fixedp.replace("<", "&lt")
                fixedp = fixedp.replace(">", "&gt")

                pfnElt.setAttribute("filetype", p[1])
                pfnElt.setAttribute("name", fixedp)
                pfnElt.setAttribute("se", p[2])

            if metadata:
                for p in self.pfns:
                    metadataElt = doc.createElement("metadata")
                    physicalElt.appendChild(metadataElt)

                    metadataElt.setAttribute("att_name", p[0])
                    metadataElt.setAttribute("att_value", p[2])

        if self.lfns:
            logicalElt = doc.createElement("logical")
            fileElt.appendChild(logicalElt)
            for l in self.lfns:
                lfnElt = doc.createElement("lfn")
                logicalElt.appendChild(lfnElt)

                lfnElt.setAttribute("name", l)
        return fileElt.toprettyxml(indent="   ")


class PoolXMLCatalog(object):
    """A Pool XML File Catalog"""

    def __init__(self, xmlfile=""):
        """PoolXMLCatalog constructor.

        Constructor takes one of the following argument types:
        xml string; list of xml strings; file name; list of file names
        """

        self.files = {}
        self.backend_file = None
        self.name = "Pool"

        # Get the dom representation of the catalog
        if xmlfile:
            if not isinstance(xmlfile, list):
                if os.path.isfile(xmlfile):
                    self.backend_file = xmlfile
                xmlfile = [xmlfile]

            for xmlf in xmlfile:
                if os.path.isfile(xmlf):
                    self.dom = xml.dom.minidom.parse(xmlf)
                else:
                    self.dom = xml.dom.minidom.parseString(xmlf)
                self.analyseCatalog(self.dom)

    def setBackend(self, fname):
        """Set the backend file name

        Sets the name of the file which will receive the contents of the
        catalog when the flush() method will be called
        """

        self.backend_file = fname

    def flush(self):
        """Flush the contents of the catalog to a file

        Flushes the contents of the catalog to a file from which
        the catalog was instanciated or which was set explicitely
        with setBackend() method
        """

        if os.path.exists(self.backend_file):
            os.rename(self.backend_file, self.backend_file + ".bak")
        with open(self.backend_file, "w") as fp:
            fp.write(self.toXML())

    def getName(self):
        """Get the catalog type name"""
        return S_OK(self.name)

    def analyseCatalog(self, dom):
        """Create the catalog from a DOM object

        Creates the contents of the catalog from the DOM XML object
        """

        catalog = dom.getElementsByTagName("POOLFILECATALOG")[0]
        pfiles = catalog.getElementsByTagName("File")
        for p in pfiles:
            guid = p.getAttribute("ID")
            pf = PoolFile(p)
            self.files[guid] = pf
            # print p.nodeName,guid

    def dump(self):
        """Dump catalog

        Dumps the contents of the catalog to the std output
        """

        for _guid, pfile in self.files.items():
            pfile.dump()

    def getFileByGuid(self, guid):
        """Get PoolFile object by GUID"""
        if guid in self.files:
            return self.files[guid]
        return None

    def getGuidByPfn(self, pfn):
        """Get GUID for a given PFN"""
        for guid, pfile in self.files.items():
            for p in pfile.pfns:
                if pfn == p[0]:
                    return guid

        return ""

    def getGuidByLfn(self, lfn):
        """Get GUID for a given LFN"""

        for guid, pfile in self.files.items():
            for l in pfile.lfns:
                if lfn == l:
                    return guid

        return ""

    def getTypeByPfn(self, pfn):
        """Get Type for a given PFN"""
        for _guid, pfile in self.files.items():
            for p in pfile.pfns:
                if pfn == p[0]:
                    return p[1]

        return ""

    def exists(self, lfn):
        """Check for the given LFN existence"""
        if self.getGuidByLfn(lfn):
            return 1
        return 0

    def getLfnsList(self):
        """Get list of LFNs in catalogue."""
        lfnsList = []
        for guid in self.files:
            lfn = self.files[guid].getLfns()
            lfnsList.append(lfn[0])

        return lfnsList

    def getLfnsByGuid(self, guid):
        """Get LFN for a given GUID"""

        lfn = ""
        if guid in self.files:
            lfns = self.files[guid].getLfns()
            lfn = lfns[0]

        if lfn:
            return S_OK(lfn)
        else:
            return S_ERROR("GUID " + guid + " not found in the catalog")

    def getPfnsByGuid(self, guid):
        """Get replicas for a given GUID"""

        result = S_OK()

        repdict = {}
        if guid in self.files:
            pfns = self.files[guid].getPfns()
            for pfn, se in pfns:
                repdict[se] = pfn
        else:
            return S_ERROR("GUID " + guid + " not found in the catalog")

        result["Replicas"] = repdict
        return result

    def getPfnsByLfn(self, lfn):
        """Get replicas for a given LFN"""

        guid = self.getGuidByLfn(lfn)
        return self.getPfnsByGuid(guid)

    def removeFileByGuid(self, guid):
        """Remove file for a given GUID"""

        for g, _pfile in self.files.items():
            if guid == g:
                del self.files[guid]

    def removeFileByLfn(self, lfn):
        """Remove file for a given LFN"""

        for guid, pfile in self.files.items():
            for l in pfile.lfns:
                if lfn == l:
                    if guid in self.files:
                        del self.files[guid]

    def addFile(self, fileTuple):
        """Add one or more files to the catalog"""

        if isinstance(fileTuple, tuple):
            files = [fileTuple]
        elif isinstance(fileTuple, list):
            files = fileTuple
        else:
            return S_ERROR("PoolXMLCatalog.addFile: Must supply a file tuple of list of tuples")

        failed = {}
        successful = {}
        for lfn, pfn, se, guid, pfnType in files:
            # print '>'*10
            # print pfnType
            pf = PoolFile()
            pf.guid = guid
            if lfn:
                pf.addLfn(lfn)
            if pfn:
                pf.addPfn(pfn, pfnType, se)

            self.files[guid] = pf
            successful[lfn] = True

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def addReplica(self, replicaTuple):
        """This adds a replica to the catalogue
        The tuple to be supplied is of the following form:

          (lfn,pfn,se,master)

        where master = True or False
        """
        if isinstance(replicaTuple, tuple):
            replicas = [replicaTuple]
        elif isinstance(replicaTuple, list):
            replicas = replicaTuple
        else:
            return S_ERROR("PoolXMLCatalog.addReplica: Must supply a replica tuple of list of tuples")

        failed = {}
        successful = {}
        for lfn, pfn, se, _master in replicas:
            guid = self.getGuidByLfn(lfn)
            if guid:
                self.files[guid].addPfn(pfn, None, se)
                successful[lfn] = True
            else:
                failed[lfn] = "LFN not found"

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def addPfnByGuid(self, lfn, guid, pfn, pfntype=None, se=None):
        """Add PFN for a given GUID - not standard"""

        if guid in self.files:
            self.files[guid].addPfn(pfn, pfntype, se)
        else:
            self.addFile([lfn, pfn, se, guid, pfntype])

    def addLfnByGuid(self, guid, lfn, pfn, se, pfntype):
        """Add LFN for a given GUID - not standard"""

        if guid in self.files:
            self.files[guid].addLfn(lfn)
        else:
            self.addFile([lfn, pfn, se, guid, pfntype])

    def toXML(self, metadata=False):
        """Convert the contents into an XML string"""

        res = """<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
<!-- Edited By PoolXMLCatalog.py -->
<!DOCTYPE POOLFILECATALOG SYSTEM "InMemory">
<POOLFILECATALOG>\n\n"""

        for _guid, pfile in self.files.items():
            res = res + pfile.toXML(metadata)

        res = res + "\n</POOLFILECATALOG>\n"
        return res



#!/usr/bin/env python
########################################################################
# File :    dirac_accounting_decode_fileid
# Author :  Adria Casajus
########################################################################
"""
Decode Accounting plot URLs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys
import pprint
from six.moves.urllib_parse import parse_qs
from six.moves.urllib import parse as urlparse

from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    from DIRAC.Core.Utilities.Plotting.FileCoding import extractRequestFromFileId

    Script.registerArgument(["URL: encoded URL of a DIRAC Accounting plot"])

    _, fileIds = Script.parseCommandLine()

    for fileId in fileIds:
        # Try to find if it's a url
        parseRes = urlparse.urlparse(fileId)
        if parseRes.query:
            queryRes = parse_qs(parseRes.query)
            if "file" in queryRes:
                fileId = queryRes["file"][0]
        # Decode
        result = extractRequestFromFileId(fileId)
        if not result["OK"]:
            gLogger.error("Could not decode fileId", "'%s', error was %s" % (fileId, result["Message"]))
            sys.exit(1)
        gLogger.notice("Decode for '%s' is:\n%s" % (fileId, pprint.pformat(result["Value"])))

    sys.exit(0)


if __name__ == "__main__":
    main()



"""
  Defines the plugin to perform evaluation on the lfn name
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

from DIRAC.Resources.Catalog.ConditionPlugins.FCConditionBasePlugin import FCConditionBasePlugin


class FilenamePlugin(FCConditionBasePlugin):
    """
    This plugin is to be used when filtering based on the LFN name
    """

    def __init__(self, conditions):
        """the condition can be any method of the python string object that can be evaluated
          as True or False:

            * endswith
            * find
            * isalnum
            * isalpha
            * isdigit
            * islower
            * isspace
            * istitle
            * isupper
            * startswith

        It should be written just like if you were calling the python call yourself.
        For example::

          Filename=startswith('/lhcb')
          Filename=istitle()

        """
        super(FilenamePlugin, self).__init__(conditions)

    def eval(self, **kwargs):
        """evaluate the parameters. The lfn argument is mandatory"""

        lfn = kwargs.get("lfn")

        if not lfn:
            return False

        evalStr = "'%s'.%s" % (lfn, self.conditions)
        try:
            ret = eval(evalStr)
            # Special case of 'find' which returns -1 if the pattern does not exist
            if self.conditions.startswith("find("):
                ret = False if ret == -1 else True

            return ret
        except Exception:
            return False



"""RabbitMQSynchronizer

  RabbitMQSynchronizer keeps the RabbitMQ user database  synchronized with the CS
  RabbitMQ user database is updated according to changes in CS.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import hostHasProperties
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNsInGroup
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForHost
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getHosts
from DIRAC.Core.Utilities.RabbitMQAdmin import getAllUsers, deleteUsers
from DIRAC.Core.Utilities.RabbitMQAdmin import setUsersPermissions, addUsersWithoutPasswords
from DIRAC.ResourceStatusSystem.Utilities import CSHelpers
from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations  # for getValidPilotGroupName()


class RabbitMQSynchronizer(object):
    def __init__(self):

        # Warm up local CS
        # I am not sure whether it is needed but
        # it was used in DIRAC.ResourceStatusSystem.Utilities.Synchronizer
        CSHelpers.warmUp()
        # only users belonging to group with this property are allowed to connect
        self._accessUserGroup = getAllowedGroupName()
        self._accessProperty = getAllowedHostProperty()  # only host with this property are allowed to connect

    def sync(self, _eventName, _params):
        """Synchronizes the internal RabbitMQ user database with the current content of CS.

        Args:
          _eventName: any value, this parameter is ignored, but needed by caller function.
          _params: - any value, this parameter is ignored, but needed by caller function.
        Returns:
          S_OK:
        Example:
           s.sync( None, None )
        """

        valid_users = getDNsInGroup(self._accessUserGroup)
        valid_hosts = getDNsForValidHosts(self._accessProperty)
        updateRabbitMQDatabase(valid_users + valid_hosts)
        return S_OK()


def getAllowedGroupName():
    """Returns name of the group of which users are allowed
       to connect to RabbitMQ server.

    Returns:
      str: group name
    """
    return Operations().getValue("Pilot/GenericPilotGroup", "")


def getAllowedHostProperty():
    """Returns property. The hosts that contain it
       are allowed to connect to RabbitMQ server.

    Returns:
      str: property name
    """
    # It should be taken from CS
    return "GenericPilot"


def getDNsForValidHosts(accessProperty):
    """Returns DN of hosts which contains accessProperty based on current
       CS settings.

    Args:
      accessProperty(str):

    Returns:
      list of hosts with accessProperty set.

    :rtype: python:list

    """

    retVal = getHosts()
    if not retVal["OK"]:
        return []
    hosts = retVal["Value"]
    DNs = []
    for host in hosts:
        if hostHasProperties(host, [accessProperty]):
            retVal = getDNForHost(host)
            if retVal["OK"]:
                DNs.extend(retVal["Value"])
            else:
                gLogger.error("Could not find a correct DN for host: %s. It will be ignored." % host)
    return DNs


def updateRabbitMQDatabase(newUsers, specialUsers=None):
    """Updates the internal user database of RabbitMQ server.
       The current user list from the database is compared to newUsers list.
       The users that are not present in the database, but are in newUsers list
       are added. The users that are present in the database, but are not in
       newUsers list are deleted. The specialUser list contains logins that
       will not be processed at all.

    Args:
      newUsers(list): user logins to be processed.
      specialUsers(list): special users that will not be processed.

    :type newUsers: python:list
    :type specialUsers: python:list

    """
    if specialUsers is None:
        # I think specialUsers should be read from CS and not taken as the argument
        # but I will leave it till we decide it.
        specialUsers = getSpecialUsersForRabbitMQDatabase()
    ret = getAllUsers()
    if not ret["OK"]:
        gLogger.error("Some problem with getting all users from RabbitMQ DB")
        return ret
    currentUsersInRabbitMQ = ret["Value"]
    # special users should not be taken into account
    currentUsersInRabbitMQ = listDifference(currentUsersInRabbitMQ, specialUsers)
    usersToRemove = listDifference(currentUsersInRabbitMQ, newUsers)
    usersToAdd = listDifference(newUsers, currentUsersInRabbitMQ)
    if usersToAdd:
        addUsersWithoutPasswords(usersToAdd)
        setUsersPermissions(usersToAdd)
    if usersToRemove:
        deleteUsers(usersToRemove)
    return S_OK()


def getSpecialUsersForRabbitMQDatabase():
    """Returns a list of special users
       that will not be processed (e.g. removed)
       while updating the RabbitMQ database.

    Returns:
      list of user logins.

    :rtype: python:list

    """
    # For a moment it is hardcoded but should be read from
    # some location in CS
    return ["admin", "dirac"]


def listDifference(list1, list2):
    """Calculates differences between two lists.
       The original order of the list is not preserved.

    :rtype: python:list
    """
    return list(set(list1) - set(list2))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# Define here a few constants useful for all the transport layer

#: Default timeout for the RPC call
DEFAULT_RPC_TIMEOUT = 600
#: Default timeout to establish a connection
DEFAULT_CONNECTION_TIMEOUT = 10

#: Default SSL Ciher accepted. Current default is for pyGSI/M2crypto compatibility
#: Can be changed with DIRAC_M2CRYPTO_SSL_CIPHERS
#: Recommandation (incompatible with pyGSI)
# pylint: disable=line-too-long
#: AES256-GCM-SHA384:AES256-SHA256:AES128-GCM-SHA256:AES128-SHA256:HIGH:MEDIUM:RSA:!3DES:!RC4:!aNULL:!eNULL:!MD5:!SEED:!IDEA:!SHA # noqa
# Cipher line should be as readable as possible, sorry pylint
# pylint: disable=line-too-long
DEFAULT_SSL_CIPHERS = "AES256-GCM-SHA384:AES256-SHA256:AES256-SHA:CAMELLIA256-SHA:AES128-GCM-SHA256:AES128-SHA256:AES128-SHA:HIGH:MEDIUM:RSA:!3DES:!RC4:!aNULL:!eNULL:!MD5:!SEED:!IDEA"  # noqa

#: Default SSL methods accepted. Current default accepts TLSv1 for pyGSI/M2crypto compatibility
#: Can be changed with DIRAC_M2CRYPTO_SSL_METHODS
#: Recommandation (incompatible with pyGSI)
# TLSv2:TLSv3
DEFAULT_SSL_METHODS = "TLSv1:TLSv2:TLSv3"



"""Test the Transformationinfo."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import OrderedDict
from contextlib import contextmanager

import pytest
from mock import MagicMock as Mock, patch

from DIRAC import S_OK, S_ERROR
import DIRAC
import DIRAC.TransformationSystem.Client.TransformationClient
import DIRAC.Resources.Catalog.FileCatalogClient
import DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient
from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo
from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo

from DIRAC.tests.Utilities.utils import MatchStringWith

__RCSID__ = "$Id$"

# pylint: disable=W0212, redefined-outer-name


@pytest.fixture
def userProxyFixture(mocker):
    """Mock UserProxy."""

    @contextmanager
    def _mockedCM(*args, **kwargs):
        try:
            yield S_OK()
        finally:
            pass

    mocker.patch("DIRAC.TransformationSystem.Utilities.TransformationInfo.UserProxy", new=_mockedCM)


@pytest.fixture
def failingUserProxyFixture(mocker):
    """Mock UserProxy."""

    @contextmanager
    def _mockedCM(*args, **kwargs):
        try:
            yield S_ERROR("Failed to set up proxy")
        finally:
            pass

    mocker.patch("DIRAC.TransformationSystem.Utilities.TransformationInfo.UserProxy", new=_mockedCM)


@pytest.fixture
def tiFixture():
    """Fixture for TransformationInfo object."""
    tMock = Mock(name="transMock", spec=DIRAC.TransformationSystem.Client.TransformationClient.TransformationClient)
    tMock.setFileStatusForTransformation = Mock(name="setFileStat")
    fcMock = Mock(name="fcMock", spec=DIRAC.Resources.Catalog.FileCatalogClient.FileCatalogClient)
    jmMock = Mock(name="jobMonMock", spec=DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient.JobMonitoringClient)
    jsucMock = Mock(name="jsuc", spec=DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient.JobStateUpdateClient)
    transInfoDict = dict(
        TransformationID=1234,
        TransformationName="TestProd12",
        Type="TestProd",
        AuthorDN="/some/cert/owner",
        AuthorGroup="Test_Prod",
    )

    tri = TransformationInfo(
        transformationID=1234, transInfoDict=transInfoDict, enabled=False, tClient=tMock, fcClient=fcMock, jobMon=jmMock
    )
    tri.log = Mock(name="LogMock")
    tri.jobStateClient = jsucMock
    return tri


@pytest.fixture
def tdFixture():
    """Fixture for tasksDicts."""
    taskDicts = [
        dict(TaskID=123, LFN="lfn123", Status="Assigned", FileID=987001, ErrorCount=9),
        dict(TaskID=124, LFN="lfn124", Status="Processed", FileID=987002, ErrorCount=8),
    ]
    return taskDicts


def test_init(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo init..........................."""
    assert isinstance(tiFixture, TransformationInfo)
    assert not tiFixture.enabled


def test_checkTasksStatus(tiFixture, tdFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo checkTasksStatus..............."""
    # error getting files
    tiFixture.tClient.getTransformationFiles.return_value = S_ERROR("nope")
    with pytest.raises(RuntimeError) as re:
        tiFixture.checkTasksStatus()
    assert "Failed to get transformation tasks: nope" in str(re)

    # success getting files
    tiFixture.tClient.getTransformationFiles.return_value = S_OK(tdFixture)
    retDict = tiFixture.checkTasksStatus()
    assert len(retDict) == 2
    assert 123 in retDict
    assert 124 in retDict
    assert len(retDict[124]) == 1
    assert "FileID" in retDict[124][0]


def test_setJob_Status(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo setJob functions..............."""
    job = Mock(spec=JobInfo)
    job.jobID = 5678
    tiFixture.enabled = False
    tiFixture._TransformationInfo__setTaskStatus = Mock()
    tiFixture._TransformationInfo__updateJobStatus = Mock()

    # setJobDone
    tiFixture.setJobDone(job)
    tiFixture._TransformationInfo__setTaskStatus.assert_not_called()
    tiFixture._TransformationInfo__updateJobStatus.assert_not_called()

    tiFixture.enabled = True
    job.status = "Done"
    tiFixture.setJobDone(job)
    tiFixture._TransformationInfo__setTaskStatus.assert_called_once_with(job, "Done")
    tiFixture._TransformationInfo__updateJobStatus.assert_not_called()

    tiFixture.enabled = True
    job.status = "Failed"
    tiFixture._TransformationInfo__setTaskStatus.reset_mock()
    tiFixture.setJobDone(job)
    tiFixture._TransformationInfo__setTaskStatus.assert_called_once_with(job, "Done")
    tiFixture._TransformationInfo__updateJobStatus.assert_called_once_with(job.jobID, "Done", "Job forced to Done")

    # setJobFailed
    tiFixture.enabled = False
    tiFixture._TransformationInfo__setTaskStatus.reset_mock()
    tiFixture._TransformationInfo__updateJobStatus.reset_mock()
    tiFixture.setJobFailed(job)
    tiFixture._TransformationInfo__setTaskStatus.assert_not_called()
    tiFixture._TransformationInfo__updateJobStatus.assert_not_called()

    tiFixture.enabled = True
    job.status = "Failed"
    tiFixture._TransformationInfo__setTaskStatus.reset_mock()
    tiFixture._TransformationInfo__updateJobStatus.reset_mock()
    tiFixture.setJobFailed(job)
    tiFixture._TransformationInfo__setTaskStatus.assert_called_once_with(job, "Failed")
    tiFixture._TransformationInfo__updateJobStatus.assert_not_called()

    tiFixture.enabled = True
    job.status = "Done"
    tiFixture._TransformationInfo__setTaskStatus.reset_mock()
    tiFixture._TransformationInfo__updateJobStatus.reset_mock()
    tiFixture.setJobFailed(job)
    tiFixture._TransformationInfo__setTaskStatus.assert_called_once_with(job, "Failed")
    tiFixture._TransformationInfo__updateJobStatus.assert_called_once_with(job.jobID, "Failed", "Job forced to Failed")


def test_setInputStatusFuncs(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo setInput functions............."""
    tiFixture._TransformationInfo__setInputStatus = Mock()
    tiFixture.enabled = False
    job = Mock(spec=JobInfo)

    tiFixture._TransformationInfo__setInputStatus.reset_mock()
    tiFixture.setInputUnused(job)
    tiFixture._TransformationInfo__setInputStatus.assert_called_once_with(job, "Unused")

    tiFixture._TransformationInfo__setInputStatus.reset_mock()
    tiFixture.setInputProcessed(job)
    tiFixture._TransformationInfo__setInputStatus.assert_called_once_with(job, "Processed")

    tiFixture._TransformationInfo__setInputStatus.reset_mock()
    tiFixture.setInputDeleted(job)
    tiFixture._TransformationInfo__setInputStatus.assert_called_once_with(job, "Deleted")

    tiFixture._TransformationInfo__setInputStatus.reset_mock()
    tiFixture.setInputMaxReset(job)
    tiFixture._TransformationInfo__setInputStatus.assert_called_once_with(job, "MaxReset")


def test_setInputStatus(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo setInputStatus................."""
    job = Mock(spec=JobInfo)
    job.inputFiles = ["dummylfn", "otherDummy"]
    status = "Unused"

    tiFixture.enabled = False
    tiFixture._TransformationInfo__setInputStatus(job, status)

    tiFixture.enabled = True
    tiFixture.tClient.setFileStatusForTransformation.return_value = S_ERROR("Failed to set")
    with pytest.raises(RuntimeError) as re:
        tiFixture._TransformationInfo__setInputStatus(job, status)
    assert "Failed updating file status" in str(re)

    tiFixture.enabled = True
    tiFixture.tClient.setFileStatusForTransformation.return_value = S_OK("All Good")
    tiFixture._TransformationInfo__setInputStatus(job, status)


def test_setTaskStatus(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo setTaskStatus.................."""
    job = Mock(spec=JobInfo)
    job.taskID = 1234
    tiFixture.tClient.setTaskStatus = Mock(return_value=S_OK("Done"))
    tiFixture._TransformationInfo__setTaskStatus(job, "Processed")
    tiFixture.tClient.setTaskStatus.assert_called_once_with(tiFixture.transName, 1234, "Processed")

    tiFixture.tClient.setTaskStatus = Mock(return_value=S_ERROR("NotDone"))
    with pytest.raises(RuntimeError) as re:
        tiFixture._TransformationInfo__setTaskStatus(job, "Processed")
    tiFixture.tClient.setTaskStatus.assert_called_once_with(tiFixture.transName, 1234, "Processed")
    assert "Failed updating task status: NotDone" in str(re)


def test_updateJobStatus(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo updateJobStatus................"""

    tiFixture.jobStateClient.setJobStatus = Mock()
    tiFixture.jobStateClient.setJobStatus.return_value = S_OK()
    tiFixture.enabled = False
    res = tiFixture._TransformationInfo__updateJobStatus(1234, "Failed", minorstatus=None)
    assert res["OK"]
    assert res["Value"] == "DisabledMode"
    tiFixture.jobStateClient.setJobStatus.assert_not_called()

    tiFixture.jobStateClient.setJobStatus.reset()
    tiFixture.jobStateClient.setJobStatus.return_value = S_OK("added record")
    tiFixture.enabled = True
    res = tiFixture._TransformationInfo__updateJobStatus(1234, "Failed", minorstatus=None)
    assert res["OK"]
    assert res["Value"] == "added record"
    tiFixture.jobStateClient.setJobStatus.assert_called_once_with(1234, "Failed", None, "DataRecoveryAgent", None, True)

    tiFixture.jobStateClient.setJobStatus.return_value = S_ERROR("Error setting job status")
    tiFixture.enabled = True
    with pytest.raises(RuntimeError) as re:
        tiFixture._TransformationInfo__updateJobStatus(1234, "Failed", minorstatus=None)
    assert "Failed to update job status" in str(re)


def test_findAllDescendants(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo findAllDescendents............."""
    tiFixture.fcClient.getFileDescendents = Mock(
        return_value=S_OK(
            {
                "Successful": {
                    "lfn1": ["lfnD1", "lfnD2"],
                    "lfnD1": ["lfnDD1", "lfnDD2"],
                }
            }
        )
    )
    descList = tiFixture._TransformationInfo__findAllDescendants(lfnList=[])
    assert set(descList) == {"lfnDD1", "lfnDD2", "lfnD1", "lfnD2"}

    tiFixture.fcClient.getFileDescendents = Mock(return_value=S_ERROR("Cannot get descendants"))
    descList = tiFixture._TransformationInfo__findAllDescendants(lfnList=[])
    assert descList == []


def test_cleanOutputs_proxyFail(tiFixture, failingUserProxyFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo cleanOutputs..................."""
    descList = ["lfnDD1", "lfnDD2", "lfnD1", "lfnD2"]

    jobInfo = Mock(spec=JobInfo)
    jobInfo.outputFiles = ["lfn1", "lfn2"]
    jobInfo.outputFileStatus = ["Exists", "Missing"]

    tiFixture.enabled = True
    tiFixture._TransformationInfo__findAllDescendants = Mock(return_value=descList)
    with pytest.raises(RuntimeError) as re:
        tiFixture.cleanOutputs(jobInfo)
    assert "Failed to get a proxy" in str(re)


def test_cleanOutputs(tiFixture, userProxyFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo cleanOutputs..................."""
    descList = ["lfnDD1", "lfnDD2", "lfnD1", "lfnD2"]

    jobInfo = Mock(spec=JobInfo)
    jobInfo.outputFiles = ["lfn1", "lfn2"]
    jobInfo.outputFileStatus = ["Exists", "Missing"]

    tiFixture.enabled = False
    tiFixture._TransformationInfo__findAllDescendants = Mock(return_value=descList)

    tiFixture.cleanOutputs(jobInfo)
    tiFixture.log.notice.assert_any_call(MatchStringWith("Would have removed these files"))
    tiFixture.log.notice.assert_any_call(MatchStringWith("lfn1"))
    for _name, args, _kwargs in tiFixture.log.notice.mock_calls:
        assert "lfn2" not in str(args)

    remMock = Mock(name="remmock")
    remMock.removeFile.return_value = S_ERROR("arg")

    tiFixture.enabled = True
    tiFixture._TransformationInfo__findAllDescendants = Mock(return_value=descList)
    with patch(
        "DIRAC.TransformationSystem.Utilities.TransformationInfo.DataManager", return_value=remMock, autospec=True
    ):
        with pytest.raises(RuntimeError) as re:
            tiFixture.cleanOutputs(jobInfo)
        assert "Failed to remove LFNs: arg" in str(re)

    remMock = Mock(name="remmock")
    remMock.removeFile.return_value = S_OK(
        {
            "Successful": {"lfn1": "OK", "lfn2": "OK"},
            "Failed": {"lfnD2": "SomeReason", "lfnD3": "SomeReason", "lfnDD2": "SomeOtherReason"},
        }
    )

    tiFixture.enabled = True
    with patch(
        "DIRAC.TransformationSystem.Utilities.TransformationInfo.DataManager", autospec=True, return_value=remMock
    ):
        tiFixture.cleanOutputs(jobInfo)
        tiFixture.log.notice.assert_any_call(MatchStringWith("Successfully removed 2 files"))

    # nothing to remove
    jobInfo = Mock(spec=JobInfo)
    jobInfo.outputFiles = []
    tiFixture._TransformationInfo__findAllDescendants = Mock(return_value=descList)
    tiFixture.cleanOutputs(jobInfo)
    tiFixture._TransformationInfo__findAllDescendants.assert_not_called()

    # nothing to remove
    jobInfo = Mock(spec=JobInfo)
    jobInfo.outputFiles = ["lfn1", "lfn2"]
    jobInfo.outputFileStatus = ["Missing", "Missing"]
    tiFixture._TransformationInfo__findAllDescendants = Mock(return_value=[])
    tiFixture.cleanOutputs(jobInfo)
    tiFixture._TransformationInfo__findAllDescendants.assert_called_once_with(jobInfo.outputFiles)


def test_getJobs(tiFixture):
    """DIRAC.TransformationSystem.Utilities.TransformationInfo getJobs........................"""
    tiFixture.jobMon.getJobs = Mock()
    tiFixture.jobMon.getJobs.side_effect = (S_OK([123, 456, 789]), S_OK([1123, 1456, 1789]))

    # All OK, just Done
    jobs, ndone, nfailed = tiFixture.getJobs(statusList=["Done"])
    attrDict = dict(Status=["Done"], JobGroup="00001234")
    tiFixture.jobMon.getJobs.assert_called_once_with(attrDict)
    assert ndone == 3
    assert nfailed == 0
    assert isinstance(jobs, OrderedDict)

    # All OK, just Done
    tiFixture.jobMon.getJobs.reset_mock()
    tiFixture.jobMon.getJobs.side_effect = (S_OK([123, 456, 789]), S_OK([1123, 1456, 1789]))
    jobs, ndone, nfailed = tiFixture.getJobs(statusList=["Failed"])
    attrDict = dict(Status=["Failed"], JobGroup="00001234")
    tiFixture.jobMon.getJobs.assert_called_once_with(attrDict)
    assert ndone == 0
    assert nfailed == 3
    assert isinstance(jobs, OrderedDict)

    # All OK, None
    tiFixture.jobMon.getJobs.reset_mock()
    tiFixture.jobMon.getJobs.side_effect = (S_OK([123, 456, 789]), S_OK([1123, 56, 89]))
    jobs, ndone, nfailed = tiFixture.getJobs(statusList=None)
    attrDict = dict(Status=["Done"], JobGroup="00001234")
    attrDict2 = dict(Status=["Failed"], JobGroup="00001234")
    tiFixture.jobMon.getJobs.assert_any_call(attrDict)
    tiFixture.jobMon.getJobs.assert_any_call(attrDict2)
    assert ndone == 3
    assert nfailed == 3
    assert isinstance(jobs, OrderedDict)
    assert [56, 89, 123, 456, 789, 1123] == list(jobs)

    # All ERROR
    tiFixture.jobMon.getJobs = Mock()
    tiFixture.jobMon.getJobs.side_effect = (S_ERROR("Not Done"), S_ERROR("Not Failed"))
    with pytest.raises(RuntimeError) as re:
        jobs, ndone, nfailed = tiFixture.getJobs(statusList=None)
    attrDict = dict(Status=["Done"], JobGroup="00001234")
    tiFixture.jobMon.getJobs.assert_called_once_with(attrDict)
    assert "Failed to get jobs" in str(re)



""" PolicyCaller

  Module used for calling policies. Its class is used for invoking
  real policies, based on the policy name.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_ERROR
from DIRAC.ResourceStatusSystem.Utilities import Utils
from DIRAC.ResourceStatusSystem.Command import CommandCaller

__RCSID__ = "$Id$"


class PolicyCaller(object):
    """
    PolicyCaller loads policies, sets commands and runs them.
    """

    def __init__(self, clients=None):
        """
        Constructor
        """

        self.cCaller = CommandCaller

        self.clients = {}
        if clients is not None:
            self.clients = clients

    def policyInvocation(self, decisionParams, policyDict):
        """
        Invokes a policy:

        1. If :attr:`policy` is None, import the policy module specified
        with :attr:`pModule` (e.g. 'DT_Policy').

          1.1. Create a policy object.

        2. Set the policy arguments (usually :attr:`granularity`,
        :attr:`name`) + :attr:`extraArgs`.

        3. If commandIn is specified (normally it is), use
        :meth:`DIRAC.ResourceStatusSystem.Command.CommandCaller.CommandCaller.setCommandObject`
        to get a command object
        """

        if "module" not in policyDict:
            return S_ERROR("Malformed policyDict %s" % policyDict)
        pModuleName = policyDict["module"]

        if "command" not in policyDict:
            return S_ERROR("Malformed policyDict %s" % policyDict)
        pCommand = policyDict["command"]

        if "args" not in policyDict:
            return S_ERROR("Malformed policyDict %s" % policyDict)
        pArgs = policyDict["args"]

        try:
            policyModule = Utils.voimport("DIRAC.ResourceStatusSystem.Policy.%s" % pModuleName)
        except ImportError:
            return S_ERROR("Unable to import DIRAC.ResourceStatusSystem.Policy.%s" % pModuleName)

        if not hasattr(policyModule, pModuleName):
            return S_ERROR("%s has no attibute %s" % (policyModule, pModuleName))

        policy = getattr(policyModule, pModuleName)()

        command = self.cCaller.commandInvocation(pCommand, pArgs, decisionParams, self.clients)
        if not command["OK"]:
            return command
        command = command["Value"]

        evaluationResult = self.policyEvaluation(policy, command)

        if evaluationResult["OK"]:
            evaluationResult["Value"]["Policy"] = policyDict

        return evaluationResult

    @staticmethod
    def policyEvaluation(policy, command):
        """
        Method that given a policy and a command objects, assigns the second one as
        a member of the first and evaluates the policy.
        """

        policy.setCommand(command)
        evaluationResult = policy.evaluate()

        return evaluationResult


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



"""
This module is used to create an appropriate object which can be used to insert records to the Monitoring system.
It always try to insert the records directly. In case of failure the monitoring client is used...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


def getDBOrClient(DB, serverName):
    """Tries to instantiate the DB object and returns it if we manage to connect to the DB,
    otherwise returns a Client of the server
    """
    from DIRAC import gLogger
    from DIRAC.Core.DISET.RPCClient import RPCClient

    try:
        database = DB()
        if database._connected:
            return database
    except Exception:
        pass

    gLogger.info("Can not connect to DB will use %s" % serverName)
    return RPCClient(serverName)


def getMonitoringDB():
    serverName = "Monitoring/Monitoring"
    MonitoringDB = None
    try:
        from DIRAC.MonitoringSystem.DB.MonitoringDB import MonitoringDB
    except Exception:
        pass
    return getDBOrClient(MonitoringDB, serverName)


monitoringDB = getMonitoringDB()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-reset-job
# Author :  Stuart Paterson
########################################################################
"""
Reset a job or list of jobs in the WMS

Example:
  $ dirac-admin-reset-job 1848
  Reset Job 1848
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job IDs"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for job in args:

        try:
            job = int(job)
        except Exception as x:
            errorList.append(("Expected integer for jobID", job))
            exitCode = 2
            continue

        result = diracAdmin.resetJob(job)
        if result["OK"]:
            print("Reset Job %s" % (job))
        else:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" PieGraph represents a pie graph

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy
import math
import time
from matplotlib.patches import Wedge, Shadow
from DIRAC.Core.Utilities.Graphs.PlotBase import PlotBase
from DIRAC.Core.Utilities.Graphs.GraphData import GraphData
from DIRAC.Core.Utilities.Graphs.GraphUtilities import *

__RCSID__ = "$Id$"


class PieGraph(PlotBase):
    def __init__(self, data, ax, prefs, *args, **kw):

        PlotBase.__init__(self, data, ax, prefs, *args, **kw)
        self.pdata = data

    def pie(self, explode=None, colors=None, autopct=None, pctdistance=0.6, shadow=False):

        start = time.time()
        labels = self.pdata.getLabels()
        if labels[0][0] == "NoLabels":
            try:
                self.pdata.initialize(key_type="string")
                self.pdata.sortLabels()
                labels = self.pdata.getLabels()
                nLabels = self.pdata.getNumberOfLabels()
                explode = [0.0] * nLabels
                if nLabels > 0:
                    explode[0] = 0.1
            except Exception as x:
                print("PieGraph Error: can not interpret data for the plot")

        # labels.reverse()
        values = [l[1] for l in labels]
        x = numpy.array(values, numpy.float64)
        self.legendData = labels

        sx = float(numpy.sum(x))
        if sx > 0:
            x = numpy.divide(x, sx)

        labels = [l[0] for l in labels]
        if explode is None:
            explode = [0] * len(x)
        assert len(x) == len(labels)
        assert len(x) == len(explode)
        plot_axis_labels = self.prefs.get("plot_axis_labels", True)

        center = 0, 0
        radius = 1.1
        theta1 = 0
        i = 0
        texts = []
        slices = []
        autotexts = []

        for frac, label, expl in zip(x, labels, explode):
            x, y = center
            theta2 = theta1 + frac
            thetam = 2 * math.pi * 0.5 * (theta1 + theta2)
            x += expl * math.cos(thetam)
            y += expl * math.sin(thetam)
            color = self.palette.getColor(label)
            w = Wedge(
                (x, y),
                radius,
                360.0 * theta1,
                360.0 * theta2,
                facecolor=color,
                lw=pixelToPoint(0.5, self.dpi),
                edgecolor="#999999",
            )
            slices.append(w)
            self.ax.add_patch(w)
            w.set_label(label)

            if shadow:
                # make sure to add a shadow after the call to
                # add_patch so the figure and transform props will be
                # set
                shad = Shadow(w, -0.02, -0.02)
                shad.set_zorder(0.9 * w.get_zorder())
                self.ax.add_patch(shad)

            if plot_axis_labels:
                if frac > 0.03:
                    xt = x + 1.05 * radius * math.cos(thetam)
                    yt = y + 1.05 * radius * math.sin(thetam)

                    thetam %= 2 * math.pi

                    if 0 < thetam and thetam < math.pi:
                        valign = "bottom"
                    elif thetam == 0 or thetam == math.pi:
                        valign = "center"
                    else:
                        valign = "top"

                    if thetam > math.pi / 2.0 and thetam < 3.0 * math.pi / 2.0:
                        halign = "right"
                    elif thetam == math.pi / 2.0 or thetam == 3.0 * math.pi / 2.0:
                        halign = "center"
                    else:
                        halign = "left"

                    t = self.ax.text(
                        xt,
                        yt,
                        label,
                        size=pixelToPoint(self.prefs["subtitle_size"], self.dpi),
                        horizontalalignment=halign,
                        verticalalignment=valign,
                    )

                    t.set_fontname(self.prefs["font"])
                    t.set_fontsize(pixelToPoint(self.prefs["text_size"], self.dpi))

                    texts.append(t)

                if autopct is not None:
                    xt = x + pctdistance * radius * math.cos(thetam)
                    yt = y + pctdistance * radius * math.sin(thetam)
                    if isinstance(autopct, str):
                        s = autopct % (100.0 * frac)
                    elif callable(autopct):
                        s = autopct(100.0 * frac)
                    else:
                        raise TypeError("autopct must be callable or a format string")

                    t = self.ax.text(xt, yt, s, horizontalalignment="center", verticalalignment="center")

                    t.set_fontname(self.prefs["font"])
                    t.set_fontsize(pixelToPoint(self.prefs["text_size"], self.dpi))

                    autotexts.append(t)

            theta1 = theta2
            i += 1

        self.legendData.reverse()

        self.ax.set_xlim((-1.25, 1.25))
        self.ax.set_ylim((-1.25, 1.25))
        self.ax.set_axis_off()

        if autopct is None:
            return slices, texts
        else:
            return slices, texts, autotexts

    min_amount = 0.1

    def getLegendData(self):

        return self.legendData

    def draw(self):

        self.ylabel = ""
        self.prefs["square_axis"] = True
        PlotBase.draw(self)

        def my_display(x):
            if x > 100 * self.min_amount:
                return "%.1f" % x + "%"
            else:
                return ""

        nLabels = self.pdata.getNumberOfLabels()
        explode = [0.0] * nLabels
        if nLabels > 0:
            explode[0] = 0.1
        self.wedges, text_labels, percent = self.pie(explode=explode, autopct=my_display)



#!/usr/bin/env python
########################################################################
# File :   dirac-configuration-dump-local-cache
# Author : Adria Casajus
########################################################################
"""
Dump DIRAC Configuration data
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys
import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.localCfg.addDefaultEntry("LogLevel", "fatal")

    fileName = ""

    def setFilename(args):
        global fileName
        fileName = args
        return DIRAC.S_OK()

    raw = False

    def setRaw(args):
        global raw
        raw = True
        return DIRAC.S_OK()

    Script.registerSwitch("f:", "file=", "Dump Configuration data into <file>", setFilename)
    Script.registerSwitch("r", "raw", "Do not make any modification to the data", setRaw)
    Script.parseCommandLine()

    from DIRAC import gConfig, gLogger

    result = gConfig.dumpCFGAsLocalCache(fileName, raw)
    if not result["OK"]:
        print("Error: %s" % result["Message"])
        sys.exit(1)

    if not fileName:
        print(result["Value"])

    sys.exit(0)


if __name__ == "__main__":
    main()



""" Unit tests for the Dirac interface module
"""
# pylint: disable=no-member, protected-access, missing-docstring
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import logging

import six
from pprint import pformat
import pytest
from mock import MagicMock, call

from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC import S_OK

logging.basicConfig()
LOG = logging.getLogger("TestDirac")
LOG.setLevel(logging.ERROR)


@pytest.fixture
def dirac():
    d = Dirac()
    d.log = MagicMock(name="Log")
    d.log.debug = d.log
    d.log.info = d.log
    d.log.notice = d.log
    d.log.verbose = d.log
    d.log.error = d.log
    d.log.warn = d.log
    return d


@pytest.fixture
def job():
    from DIRAC.Interfaces.API.Job import Job

    job = Job(stdout="printer", stderr="/dev/null")
    job.setInputSandbox(["LFN:/vo/user/i/initial/important.tar.gz", "/abspath/absfile.xml", "file_in_pwd.xml"])
    return job


@pytest.fixture
def osmock():
    os = MagicMock(return_value=False, name="OS")
    os.environ = dict()
    os.environ["DIRAC"] = "/root/dirac"

    def expandMock(*args, **kwargs):
        if "DIRACROOT" in os.environ and "DIRACROOT" in args[0]:
            return args[0].replace("$DIRACROOT", os.environ["DIRACROOT"])
        return args[0]

    def existsMock(*args, **kwargs):
        print("exists", args)
        if any(f in args[0] for f in ("jobDescription", "/pwd/file_in_pwd.xml")):
            return True
        if "absfile" in args[0] and args[0].startswith("/abspath"):
            return True
        return False

    def joinMock(*args):
        path = "/"
        for l in args:
            path += l.strip("/") + "/"
        return path.rstrip("/")

    os.getcwd.side_effect = ["/pwd/"] + ["/pwd/tempFolder/"] * 100

    os.path.expandvars.side_effect = expandMock
    os.path.exists.side_effect = existsMock
    os.path.isabs.side_effect = lambda x: x.startswith("/")
    os.path.isdir.return_value = False
    os.path.join.side_effect = joinMock
    os.path.basename.side_effect = lambda x: x.rsplit("/", 1)[-1]
    return os


@pytest.fixture
def confMock():
    gConf = MagicMock(name="gConfig")

    def getVal(*args, **kwargs):
        if "/LocalSite/Root" in args[0]:
            return "/root/dirac"
        if len(args) == 2:
            return args[1]
        return "defaultValue"

    gConf.getValue.side_effect = getVal
    return gConf


def test_basicJob(dirac):
    jdl = "Parameter=Value;Parameter2=Value2"
    ret = dirac._Dirac__getJDLParameters(jdl)
    assert ret["OK"]
    assert "Parameter" in ret["Value"]
    assert ret["Value"]["Parameter"] == "Value"
    assert "Parameter2" in ret["Value"]
    assert ret["Value"]["Parameter2"] == "Value2"


def test_JobJob(dirac, job):
    ret = dirac._Dirac__getJDLParameters(job)
    assert ret["OK"]
    assert ret["Value"]["StdOutput"] == "printer"
    assert ret["Value"]["StdError"] == "/dev/null"


def test_runLocal(dirac, job, mocker, osmock, confMock):
    mocker.patch("DIRAC.Interfaces.API.Dirac.os", new=osmock)
    mocker.patch("DIRAC.Interfaces.API.Dirac.tarfile", new=MagicMock(return_value=False))
    if six.PY2:
        mocker.patch("__builtin__.open")
    else:
        mocker.patch("builtins.open")
    mocker.patch("DIRAC.Interfaces.API.Dirac.gConfig", new=confMock)
    tempMock = mocker.patch("DIRAC.Interfaces.API.Dirac.tempfile")
    tempMock.mkdtemp.return_value = "tempFolder"
    shMock = mocker.patch("DIRAC.Interfaces.API.Dirac.shutil", new=MagicMock(name="Shutil"))
    sysMock = mocker.patch("DIRAC.Interfaces.API.Dirac.systemCall")
    sysMock.return_value = S_OK([0, "No output", "No errors"])
    dirac.getFile = MagicMock(return_value=S_OK())
    ret = dirac.runLocal(job)
    LOG.info("dirac log calls: %s", dirac.log.call_args_list)
    LOG.info("CallStack: %s", pformat(ret.get("CallStack", {})))
    try:
        assert sysMock.call_args_list[0][1]["cmdSeq"] == ["dirac-jobexec", "jobDescription.xml", "-o", "LogLevel=DEBUG"]
    except AssertionError:
        assert sysMock.call_args_list[0][1]["cmdSeq"] == [
            "/root/dirac/scripts/dirac-jobexec",
            "jobDescription.xml",
            "-o",
            "LogLevel=DEBUG",
        ]
    assert ret.get("Message", None) is None
    assert ret["OK"]
    assert call("/abspath/absfile.xml", "/pwd/tempFolder/") in shMock.copy.call_args_list
    assert call("/pwd/file_in_pwd.xml", "/pwd/tempFolder/") in shMock.copy.call_args_list
    assert call("/vo/user/i/initial/important.tar.gz") in dirac.getFile.call_args_list



"""
    TornadoBaseClient contains all the low-levels functionalities and initilization methods
    It must be instantiated from :py:class:`~DIRAC.Core.Tornado.Client.TornadoClient`

    Requests library manage itself retry when connection failed, so the __nbOfRetry attribute is removed from DIRAC
    (For each URL requests manage retries himself, if it still fail, we try next url)
    KeepAlive lapse is also removed because managed by request,
    see https://requests.readthedocs.io/en/latest/user/advanced/#keep-alive

    If necessary this class can be modified to define number of retry in requests, documentation does not give
    lot of informations but you can see this simple solution from StackOverflow.
    After some tests request seems to retry 3 times by default.
    https://stackoverflow.com/questions/15431044/can-i-set-max-retries-for-requests-request

    .. warning::
      If you use your own certificates, it's like in dips, please take a look at :ref:`using_own_CA`

    .. warning::
      Lots of method are copy-paste from :py:class:`~DIRAC.Core.DISET.private.BaseClient`.
      And some methods are copy-paste AND modifications, for now it permit to fully separate DISET and HTTPS.


"""

# pylint: disable=broad-except

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from io import open
import errno
import os
import requests
import six
import tempfile
from six.moves import http_client


from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import skipCACheck
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import findDefaultGroupForDN
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceURLs, getGatewayURLs

from DIRAC.Core.DISET.ThreadConfig import ThreadConfig
from DIRAC.Core.Security import Locations
from DIRAC.Core.Utilities import Network
from DIRAC.Core.Utilities.JEncode import decode, encode


# TODO CHRIS: refactor all the messy `discover` methods
# I do not do it now because I want first to decide
# whether we go with code copy of fatorization


class TornadoBaseClient(object):
    """
    This class contain initialization method and all utilities method used for RPC
    """

    __threadConfig = ThreadConfig()
    VAL_EXTRA_CREDENTIALS_HOST = "hosts"

    KW_USE_CERTIFICATES = "useCertificates"
    KW_EXTRA_CREDENTIALS = "extraCredentials"
    KW_TIMEOUT = "timeout"
    KW_SETUP = "setup"
    KW_VO = "VO"
    KW_DELEGATED_DN = "delegatedDN"
    KW_DELEGATED_GROUP = "delegatedGroup"
    KW_IGNORE_GATEWAYS = "ignoreGateways"
    KW_PROXY_LOCATION = "proxyLocation"
    KW_PROXY_STRING = "proxyString"
    KW_PROXY_CHAIN = "proxyChain"
    KW_SKIP_CA_CHECK = "skipCACheck"
    KW_KEEP_ALIVE_LAPSE = "keepAliveLapse"

    def __init__(self, serviceName, **kwargs):
        """
        :param serviceName: URL of the service (proper uri or just System/Component)
        :param useCertificates: If set to True, use the server certificate
        :param extraCredentials:
        :param timeout: Timeout of the call (default 600 s)
        :param setup: Specify the Setup
        :param VO: Specify the VO
        :param delegatedDN: Not clear what it can be used for.
        :param delegatedGroup: Not clear what it can be used for.
        :param ignoreGateways: Ignore the DIRAC Gatways settings
        :param proxyLocation: Specify the location of the proxy
        :param proxyString: Specify the proxy string
        :param proxyChain: Specify the proxy chain
        :param skipCACheck: Do not check the CA
        :param keepAliveLapse: Duration for keepAliveLapse (heartbeat like)  (now managed by requests)
        """

        if not isinstance(serviceName, six.string_types):
            raise TypeError(
                "Service name expected to be a string. Received %s type %s" % (str(serviceName), type(serviceName))
            )

        self._destinationSrv = serviceName
        self._serviceName = serviceName
        self.__ca_location = False

        self.kwargs = kwargs
        self.__useCertificates = None
        # The CS useServerCertificate option can be overridden by explicit argument
        self.__forceUseCertificates = self.kwargs.get(self.KW_USE_CERTIFICATES)
        self.__initStatus = S_OK()
        self.__idDict = {}
        self.__extraCredentials = ""
        # by default we always have 1 url for example:
        # RPCClient('dips://volhcb38.cern.ch:9162/Framework/SystemAdministrator')
        self.__nbOfUrls = 1
        self.__bannedUrls = []

        # For pylint...
        self.setup = None
        self.vo = None
        self.serviceURL = None

        for initFunc in (
            self.__discoverTimeout,
            self.__discoverSetup,
            self.__discoverVO,
            self.__discoverCredentialsToUse,
            self.__discoverExtraCredentials,
            self.__discoverURL,
        ):

            result = initFunc()
            if not result["OK"] and self.__initStatus["OK"]:
                self.__initStatus = result

    def __discoverSetup(self):
        """Discover which setup to use and stores it in self.setup
        The setup is looked for:
           * kwargs of the constructor (see KW_SETUP)
           * in the CS /DIRAC/Setup
           * default to 'Test'
        """
        if self.KW_SETUP in self.kwargs and self.kwargs[self.KW_SETUP]:
            self.setup = str(self.kwargs[self.KW_SETUP])
        else:
            self.setup = self.__threadConfig.getSetup()
            if not self.setup:
                self.setup = gConfig.getValue("/DIRAC/Setup", "Test")
        return S_OK()

    def __discoverURL(self):
        """Calculate the final URL. It is called at initialization and in connect in case of issue

          It sets:
            * self.serviceURL: the url (dips) selected as target using __findServiceURL
            * self.__URLTuple: a split of serviceURL obtained by Network.splitURL
            * self._serviceName: the last part of URLTuple (typically System/Component)

        WARNING: COPY PASTE FROM BaseClient
        """
        # Calculate final URL
        try:
            result = self.__findServiceURL()
        except Exception as e:
            return S_ERROR(repr(e))
        if not result["OK"]:
            return result
        self.serviceURL = result["Value"]
        retVal = Network.splitURL(self.serviceURL)
        if not retVal["OK"]:
            return retVal
        self.__URLTuple = retVal["Value"]
        self._serviceName = self.__URLTuple[-1]
        res = gConfig.getOptionsDict("/DIRAC/ConnConf/%s:%s" % self.__URLTuple[1:3])
        if res["OK"]:
            opts = res["Value"]
            for k in opts:
                if k not in self.kwargs:
                    self.kwargs[k] = opts[k]
        return S_OK()

    def __discoverVO(self):
        """Discover which VO to use and stores it in self.vo
        The VO is looked for:
           * kwargs of the constructor (see KW_VO)
           * in the CS /DIRAC/VirtualOrganization
           * default to 'unknown'

        WARNING: COPY/PASTE FROM Core/Diset/private/BaseClient FOR NOW
        """
        if self.KW_VO in self.kwargs and self.kwargs[self.KW_VO]:
            self.vo = str(self.kwargs[self.KW_VO])
        else:
            self.vo = gConfig.getValue("/DIRAC/VirtualOrganization", "unknown")
        return S_OK()

    def __discoverCredentialsToUse(self):
        """Discovers which credentials to use for connection.
        * Server certificate:
          -> If KW_USE_CERTIFICATES in kwargs, sets it in self.__useCertificates
          -> If not, check gConfig.useServerCertificate(), and sets it in self.__useCertificates
              and kwargs[KW_USE_CERTIFICATES]
        * Certification Authorities check:
           -> if KW_SKIP_CA_CHECK is not in kwargs and we are using the certificates,
                set KW_SKIP_CA_CHECK to false in kwargs
           -> if KW_SKIP_CA_CHECK is not in kwargs and we are not using the certificate, check the skipCACheck
        * Proxy Chain

        WARNING: MOSTLY COPY/PASTE FROM Core/Diset/private/BaseClient

        """
        # Use certificates?
        if self.KW_USE_CERTIFICATES in self.kwargs:
            self.__useCertificates = self.kwargs[self.KW_USE_CERTIFICATES]
        else:
            self.__useCertificates = gConfig.useServerCertificate()
            self.kwargs[self.KW_USE_CERTIFICATES] = self.__useCertificates
        if self.KW_SKIP_CA_CHECK not in self.kwargs:
            if self.__useCertificates:
                self.kwargs[self.KW_SKIP_CA_CHECK] = False
            else:
                self.kwargs[self.KW_SKIP_CA_CHECK] = skipCACheck()

        # Rewrite a little bit from here: don't need the proxy string, we use the file
        if self.KW_PROXY_CHAIN in self.kwargs:
            try:
                self.kwargs[self.KW_PROXY_STRING] = self.kwargs[self.KW_PROXY_CHAIN].dumpAllToString()["Value"]
                del self.kwargs[self.KW_PROXY_CHAIN]
            except Exception:
                return S_ERROR("Invalid proxy chain specified on instantiation")

        # ==== REWRITED FROM HERE ====

        # For certs always check CA's. For clients skipServerIdentityCheck

        return S_OK()

    def __discoverExtraCredentials(self):
        """Add extra credentials informations.
            * self.__extraCredentials
              -> if KW_EXTRA_CREDENTIALS in kwargs, we set it
              -> Otherwise, if we use the server certificate, we set it to VAL_EXTRA_CREDENTIALS_HOST
              -> If we have a delegation (see bellow), we set it to (delegatedDN, delegatedGroup)
              -> otherwise it is an empty string
            * delegation:
              -> if KW_DELEGATED_DN in kwargs, or delegatedDN in threadConfig, put in in self.kwargs
              -> If we have a delegated DN but not group, we find the corresponding group in the CS

        WARNING: COPY/PASTE FROM Core/Diset/private/BaseClient
        """
        # which extra credentials to use?
        if self.__useCertificates:
            self.__extraCredentials = self.VAL_EXTRA_CREDENTIALS_HOST
        else:
            self.__extraCredentials = ""
        if self.KW_EXTRA_CREDENTIALS in self.kwargs:
            self.__extraCredentials = self.kwargs[self.KW_EXTRA_CREDENTIALS]
        # Are we delegating something?
        delegatedDN, delegatedGroup = self.__threadConfig.getID()
        if self.KW_DELEGATED_DN in self.kwargs and self.kwargs[self.KW_DELEGATED_DN]:
            delegatedDN = self.kwargs[self.KW_DELEGATED_DN]
        elif delegatedDN:
            self.kwargs[self.KW_DELEGATED_DN] = delegatedDN
        if self.KW_DELEGATED_GROUP in self.kwargs and self.kwargs[self.KW_DELEGATED_GROUP]:
            delegatedGroup = self.kwargs[self.KW_DELEGATED_GROUP]
        elif delegatedGroup:
            self.kwargs[self.KW_DELEGATED_GROUP] = delegatedGroup
        if delegatedDN:
            if not delegatedGroup:
                result = findDefaultGroupForDN(self.kwargs[self.KW_DELEGATED_DN])
                if not result["OK"]:
                    return result
            self.__extraCredentials = (delegatedDN, delegatedGroup)
        return S_OK()

    def __discoverTimeout(self):
        """Discover which timeout to use and stores it in self.timeout
        The timeout can be specified kwargs of the constructor (see KW_TIMEOUT),
        with a minimum of 120 seconds.
        If unspecified, the timeout will be 600 seconds.
        The value is set in self.timeout, as well as in self.kwargs[KW_TIMEOUT]

        WARNING: COPY/PASTE FROM Core/Diset/private/BaseClient
        """
        if self.KW_TIMEOUT in self.kwargs:
            self.timeout = self.kwargs[self.KW_TIMEOUT]
        else:
            self.timeout = False
        if self.timeout:
            self.timeout = max(120, self.timeout)
        else:
            self.timeout = 600
        self.kwargs[self.KW_TIMEOUT] = self.timeout
        return S_OK()

    def __findServiceURL(self):
        """
        Discovers the URL of a service, taking into account gateways, multiple URLs, banned URLs


        If the site on which we run is configured to use gateways (/DIRAC/Gateways/<siteName>),
        these URLs will be used. To ignore the gateway, it is possible to set KW_IGNORE_GATEWAYS
        to False in kwargs.

        If self._destinationSrv (given as constructor attribute) is a properly formed URL,
        we just return this one. If we have to use a gateway, we just replace the server name in the url.

        The list of URLs defined in the CS (<System>/URLs/<Component>) is randomized

        This method also sets some attributes:
          * self.__nbOfUrls = number of URLs
          * self.__nbOfRetry removed in HTTPS (Managed by requests)
          * self.__bannedUrls is reinitialized if all the URLs are banned

        :return: the selected URL

        WARNING (Mostly) COPY PASTE FROM BaseClient (protocols list is changed to https)

        """
        if not self.__initStatus["OK"]:
            return self.__initStatus

        # Load the Gateways URLs for the current site Name
        gatewayURL = False
        if not self.kwargs.get(self.KW_IGNORE_GATEWAYS):
            gatewayURLs = getGatewayURLs()
            if gatewayURLs:
                gatewayURL = "/".join(gatewayURLs[0].split("/")[:3])

        # If what was given as constructor attribute is a properly formed URL,
        # we just return this one.
        # If we have to use a gateway, we just replace the server name in it
        if self._destinationSrv.startswith("https://"):
            gLogger.debug("Already given a valid url", self._destinationSrv)
            if not gatewayURL:
                return S_OK(self._destinationSrv)
            gLogger.debug("Reconstructing given URL to pass through gateway")
            path = "/".join(self._destinationSrv.split("/")[3:])
            finalURL = "%s/%s" % (gatewayURL, path)
            gLogger.debug("Gateway URL conversion:\n %s -> %s" % (self._destinationSrv, finalURL))
            return S_OK(finalURL)

        if gatewayURL:
            gLogger.debug("Using gateway", gatewayURL)
            return S_OK("%s/%s" % (gatewayURL, self._destinationSrv))

        # If nor url is given as constructor, we extract the list of URLs from the CS (System/URLs/Component)
        try:
            # We randomize the list, and add at the end the failover URLs (System/FailoverURLs/Component)
            urlsList = getServiceURLs(self._destinationSrv, setup=self.setup, failover=True)
        except Exception as e:
            return S_ERROR("Cannot get URL for %s in setup %s: %s" % (self._destinationSrv, self.setup, repr(e)))
        if not urlsList:
            return S_ERROR("URL for service %s not found" % self._destinationSrv)

        self.__nbOfUrls = len(urlsList)
        # __nbOfRetry removed in HTTPS (managed by requests)
        if self.__nbOfUrls == len(self.__bannedUrls):
            self.__bannedUrls = []  # retry all urls
            gLogger.debug("Retrying again all URLs")

        if self.__bannedUrls and len(urlsList) > 1:
            # we have host which is not accessible. We remove that host from the list.
            # We only remove if we have more than one instance
            for i in self.__bannedUrls:
                gLogger.debug("Removing banned URL", "%s" % i)
                urlsList.remove(i)

        sURL = urlsList[0]

        # If we have banned URLs, and several URLs at disposals, we make sure that the selected sURL
        # is not on a host which is banned. If it is, we take the next one in the list using __selectUrl

        if self.__bannedUrls and self.__nbOfUrls > 2:  # when we have multiple services then we can
            # have a situation when two services are running on the same machine with different ports...
            retVal = Network.splitURL(sURL)
            nexturl = None
            if retVal["OK"]:
                nexturl = retVal["Value"]

                found = False
                for i in self.__bannedUrls:
                    retVal = Network.splitURL(i)
                    if retVal["OK"]:
                        bannedurl = retVal["Value"]
                    else:
                        break
                    # We found a banned URL on the same host as the one we are running on
                    if nexturl[1] == bannedurl[1]:
                        found = True
                        break
                if found:
                    nexturl = self.__selectUrl(nexturl, urlsList[1:])
                    if nexturl:  # an url found which is in different host
                        sURL = nexturl
        gLogger.debug("Discovering URL for service", "%s -> %s" % (self._destinationSrv, sURL))
        return S_OK(sURL)

    def __selectUrl(self, notselect, urls):
        """In case when multiple services are running in the same host, a new url has to be in a different host
        Note: If we do not have different host we will use the selected url...

        :param notselect: URL that should NOT be selected
        :param urls: list of potential URLs

        :return: selected URL

        WARNING: COPY/PASTE FROM Core/Diset/private/BaseClient
        """
        url = None
        for i in urls:
            retVal = Network.splitURL(i)
            if retVal["OK"]:
                if retVal["Value"][1] != notselect[1]:  # the hots are different
                    url = i
                    break
                else:
                    gLogger.error(retVal["Message"])
        return url

    def getServiceName(self):
        """
        Returns the name of the service, if you had given a url at init, returns the URL.
        """
        return self._serviceName

    def getDestinationService(self):
        """
        Returns the url the service.
        """
        urls = getServiceURLs(self._serviceName)
        return urls[0] if urls else ""

    def _getBaseStub(self):
        """Returns a tuple with (self._destinationSrv, newKwargs)
        self._destinationSrv is what was given as first parameter of the init serviceName

        newKwargs is an updated copy of kwargs:
          * if set, we remove the useCertificates (KW_USE_CERTIFICATES) in newKwargs

        This method is just used to return information in case of error in the InnerRPCClient

        WARNING: COPY/PASTE FROM Core/Diset/private/BaseClient
        """
        newKwargs = dict(self.kwargs)
        # Remove useCertificates as the forwarder of the call will have to
        # independently decide whether to use their cert or not anyway.
        if "useCertificates" in newKwargs:
            del newKwargs["useCertificates"]
        return (self._destinationSrv, newKwargs)

    def _request(self, retry=0, outputFile=None, **kwargs):
        """
        Sends the request to server

        :param retry: internal parameters for recursive call. TODO: remove ?
        :param outputFile: (default None) path to a file where to store the received data.
                          If set, the server response will be streamed for optimization
                          purposes, and the response data will not go through the
                          JDecode process
        :param **kwargs: Any argument there is used as a post parameter. They are detailed bellow.
        :param method: (mandatory) name of the distant method
        :param args: (mandatory) json serialized list of argument for the procedure



        :returns: The received data. If outputFile is set, return always S_OK

        """

        # Adding some informations to send
        if self.__extraCredentials:
            kwargs[self.KW_EXTRA_CREDENTIALS] = encode(self.__extraCredentials)
        kwargs["clientVO"] = self.vo
        kwargs["clientSetup"] = self.setup

        # Getting URL
        url = self.__findServiceURL()
        if not url["OK"]:
            return url
        url = url["Value"]

        # Getting CA file (or skip verification)
        verify = not self.kwargs.get(self.KW_SKIP_CA_CHECK)
        if verify:
            cafile = Locations.getCAsLocation()
            if not cafile:
                gLogger.error("No CAs found!")
                return S_ERROR("No CAs found!")
            verify = self.__ca_location

        # getting certificate
        # Do we use the server certificate ?
        if self.kwargs[self.KW_USE_CERTIFICATES]:
            cert = Locations.getHostCertificateAndKeyLocation()
        elif self.kwargs.get(self.KW_PROXY_STRING):
            tmpHandle, cert = tempfile.mkstemp()
            fp = os.fdopen(tmpHandle, "wb")
            fp.write(self.kwargs[self.KW_PROXY_STRING])
            fp.close()

        # CHRIS 04.02.21
        # TODO: add proxyLocation check ?
        else:
            cert = Locations.getProxyLocation()
            if not cert:
                gLogger.error("No proxy found")
                return S_ERROR("No proxy found")

        # We have a try/except for all the exceptions
        # whose default behavior is to try again,
        # maybe to different server
        try:
            # And we have a second block to handle specific exceptions
            # which makes it not worth retrying
            try:
                rawText = None

                # Default case, just return the result
                if not outputFile:
                    call = requests.post(url, data=kwargs, timeout=self.timeout, verify=verify, cert=cert)
                    # raising the exception for status here
                    # means essentialy that we are losing here the information of what is returned by the server
                    # as error message, since it is not passed to the exception
                    # However, we can store the text and return it raw as an error,
                    # since there is no guarantee that it is any JEncoded text
                    # Note that we would get an exception only if there is an exception on the server side which
                    # is not handled.
                    # Any standard S_ERROR will be transfered as an S_ERROR with a correct code.
                    rawText = call.text
                    call.raise_for_status()
                    return decode(rawText)[0]
                else:
                    # Instruct the server not to encode the response
                    kwargs["rawContent"] = True

                    rawText = None
                    # Stream download
                    # https://requests.readthedocs.io/en/latest/user/advanced/#body-content-workflow
                    with requests.post(
                        url, data=kwargs, timeout=self.timeout, verify=verify, cert=cert, stream=True
                    ) as r:
                        rawText = r.text
                        r.raise_for_status()

                        with open(outputFile, "wb") as f:
                            for chunk in r.iter_content(4096):
                                # if chunk:  # filter out keep-alive new chuncks
                                f.write(chunk)

                        return S_OK()

            # Some HTTPError are not worth retrying
            except requests.exceptions.HTTPError as e:
                status_code = e.response.status_code
                if status_code == http_client.NOT_IMPLEMENTED:
                    return S_ERROR(errno.ENOSYS, "%s is not implemented" % kwargs.get("method"))
                elif status_code in (http_client.FORBIDDEN, http_client.UNAUTHORIZED):
                    return S_ERROR(errno.EACCES, "No access to %s" % url)

                # if it is something else, retry
                raise

        # Whatever exception we have here, we deem worth retrying
        except Exception as e:
            # CHRIS TODO review this part: retry logic is fishy
            # self.__bannedUrls is emptied in findServiceURLs
            if url not in self.__bannedUrls:
                self.__bannedUrls += [url]
            if retry < self.__nbOfUrls - 1:
                self._request(retry=retry + 1, outputFile=outputFile, **kwargs)

            errStr = "%s: %s" % (str(e), rawText)
            return S_ERROR(errStr)


# --- TODO ----
# Rewrite this method if needed:
#  /Core/DISET/private/BaseClient.py
# __delegateCredentials



########################################################################
# File :   ProxyGeneration.py
# Author : Adrian Casajus
########################################################################
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import sys
from prompt_toolkit import prompt
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.NTP import getClockDeviation

__RCSID__ = "$Id$"


class CLIParams(object):

    proxyLifeTime = 86400
    diracGroup = False
    proxyStrength = 1024
    limitedProxy = False
    strict = False
    summary = False
    certLoc = False
    keyLoc = False
    proxyLoc = False
    checkWithCS = True
    stdinPasswd = False
    userPasswd = ""
    checkClock = True
    embedDefaultGroup = True
    rfc = True

    def setProxyLifeTime(self, arg):
        """Set proxy lifetime

        :param str arg: arguments

        :return: S_OK()/S_ERROR()
        """
        try:
            fields = [f.strip() for f in arg.split(":")]
            self.proxyLifeTime = int(fields[0]) * 3600 + int(fields[1]) * 60
        except Exception:
            gLogger.error("Can't parse time! Is it a HH:MM?", arg)
            return S_ERROR("Can't parse time argument")
        return S_OK()

    def setRFC(self, _arg):
        """Set RFC

        :param _arg: unuse

        :return: S_OK()
        """
        self.rfc = True
        return S_OK()

    def setNoRFC(self, _arg):
        """Unset RFC

        :param _arg: unuse

        :return: S_OK()
        """
        self.rfc = False
        return S_OK()

    def setProxyRemainingSecs(self, arg):
        """Set proxy lifetime

        :param int arg: lifetime in seconds

        :return: S_OK()
        """
        self.proxyLifeTime = int(arg)
        return S_OK()

    def getProxyLifeTime(self):
        """Get proxy lifetime

        :return: str
        """
        hours = int(self.proxyLifeTime / 3600)
        mins = int(self.proxyLifeTime / 60 - hours * 60)
        return "%s:%s" % (hours, mins)

    def getProxyRemainingSecs(self):
        """Get proxy livetime

        :return: int
        """
        return self.proxyLifeTime

    def setDIRACGroup(self, arg):
        """Set DIRAC group

        :param str arg: arguments

        :return: S_OK()
        """
        self.diracGroup = arg
        return S_OK()

    def getDIRACGroup(self):
        """Get DIRAC group

        :return: str
        """
        return self.diracGroup

    def setProxyStrength(self, arg):
        """Get proxy strength

        :return: S_OK()
        """
        try:
            self.proxyStrength = int(arg)
        except Exception:
            gLogger.error("Can't parse bits! Is it a number?", "%s" % arg)
            return S_ERROR("Can't parse strength argument")
        return S_OK()

    def setProxyLimited(self, _arg):
        """Set proxy limited

        :param _arg: unuse

        :return: str
        """
        self.limitedProxy = True
        return S_OK()

    def setSummary(self, _arg):
        """Set proxy limited

        :param _arg: unuse

        :return: str
        """
        gLogger.info("Enabling summary output")
        self.summary = True
        return S_OK()

    def setCertLocation(self, arg):
        """Set certificate path

        :param str arg: certificate path

        :return: S_OK()
        """
        self.certLoc = arg
        return S_OK()

    def setKeyLocation(self, arg):
        """Set key path

        :param str arg: key path

        :return: S_OK()
        """
        self.keyLoc = arg
        return S_OK()

    def setProxyLocation(self, arg):
        """Set proxy path

        :param str arg: proxy path

        :return: S_OK()
        """
        self.proxyLoc = arg
        return S_OK()

    def setDisableCSCheck(self, _arg):
        """Disable CS check

        :param _arg: unuse

        :return: S_OK()
        """
        self.checkWithCS = False
        return S_OK()

    def setStdinPasswd(self, _arg):
        """Set stdin passwd

        :param _arg: unuse

        :return: S_OK()
        """
        self.stdinPasswd = True
        return S_OK()

    def setStrict(self, _arg):
        """Set strict

        :param _arg: unuse

        :return: S_OK()
        """
        self.strict = True
        return S_OK()

    def showVersion(self, _arg):
        """Show version

        :param _arg: unuse

        :return: S_OK()
        """
        gLogger.always("Version: %s" % __RCSID__)
        sys.exit(0)
        return S_OK()

    def disableClockCheck(self, _arg):
        """Disable clock check

        :param _arg: unuse

        :return: S_OK()
        """
        self.checkClock = False
        return S_OK()

    def registerCLISwitches(self):
        """Register CLI switches"""
        Script.registerSwitch(
            "v:", "valid=", "Valid HH:MM for the proxy. By default is 24 hours", self.setProxyLifeTime
        )
        Script.registerSwitch("g:", "group=", "DIRAC Group to embed in the proxy", self.setDIRACGroup)
        Script.registerSwitch("b:", "strength=", "Set the proxy strength in bytes", self.setProxyStrength)
        Script.registerSwitch("l", "limited", "Generate a limited proxy", self.setProxyLimited)
        Script.registerSwitch("t", "strict", "Fail on each error. Treat warnings as errors.", self.setStrict)
        Script.registerSwitch("S", "summary", "Enable summary output when generating proxy", self.setSummary)
        Script.registerSwitch("C:", "Cert=", "File to use as user certificate", self.setCertLocation)
        Script.registerSwitch("K:", "Key=", "File to use as user key", self.setKeyLocation)
        Script.registerSwitch("u:", "out=", "File to write as proxy", self.setProxyLocation)
        Script.registerSwitch("x", "nocs", "Disable CS check", self.setDisableCSCheck)
        Script.registerSwitch("p", "pwstdin", "Get passwd from stdin", self.setStdinPasswd)
        Script.registerSwitch("i", "version", "Print version", self.showVersion)
        Script.registerSwitch("j", "noclockcheck", "Disable checking if time is ok", self.disableClockCheck)
        Script.registerSwitch("r", "rfc", "Create an RFC proxy, true by default, deprecated flag", self.setRFC)
        Script.registerSwitch("L", "legacy", "Create a legacy non-RFC proxy", self.setNoRFC)


from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Security import Locations


def generateProxy(params):
    """Generate proxy

    :param params: parameters

    :return: S_OK()/S_ERROR()
    """
    if params.checkClock:
        result = getClockDeviation()
        if result["OK"]:
            deviation = result["Value"]
            if deviation > 600:
                gLogger.error("Your host clock seems to be off by more than TEN MINUTES! Thats really bad.")
                gLogger.error("We're cowardly refusing to generate a proxy. Please fix your system time")
                sys.exit(1)
            elif deviation > 180:
                gLogger.error("Your host clock seems to be off by more than THREE minutes! Thats bad.")
                gLogger.notice("We'll generate the proxy but please fix your system time")
            elif deviation > 60:
                gLogger.error("Your host clock seems to be off by more than a minute! Thats not good.")
                gLogger.notice("We'll generate the proxy but please fix your system time")

    certLoc = params.certLoc
    keyLoc = params.keyLoc
    if not certLoc or not keyLoc:
        cakLoc = Locations.getCertificateAndKeyLocation()
        if not cakLoc:
            return S_ERROR("Can't find user certificate and key")
        if not certLoc:
            certLoc = cakLoc[0]
        if not keyLoc:
            keyLoc = cakLoc[1]
    params.certLoc = certLoc
    params.keyLoc = keyLoc

    # Load password
    testChain = X509Chain()
    retVal = testChain.loadChainFromFile(params.certLoc)
    if not retVal["OK"]:
        return S_ERROR("Cannot load certificate %s: %s" % (params.certLoc, retVal["Message"]))
    timeLeft = int(testChain.getRemainingSecs()["Value"] / 86400)
    if timeLeft < 30:
        gLogger.notice("\nYour certificate will expire in %d days. Please renew it!\n" % timeLeft)

    # First try reading the key from the file
    retVal = testChain.loadKeyFromFile(params.keyLoc, password=params.userPasswd)  # XXX why so commented?
    if not retVal["OK"]:
        if params.stdinPasswd:
            userPasswd = sys.stdin.readline().strip("\n")
        else:
            try:
                userPasswd = prompt(u"Enter Certificate password: ", is_password=True)
            except KeyboardInterrupt:
                return S_ERROR("Caught KeyboardInterrupt, exiting...")
        params.userPasswd = userPasswd

    # Find location
    proxyLoc = params.proxyLoc
    if not proxyLoc:
        proxyLoc = Locations.getDefaultProxyLocation()

    chain = X509Chain()
    # Load user cert and key
    retVal = chain.loadChainFromFile(certLoc)
    if not retVal["OK"]:
        gLogger.warn(retVal["Message"])
        return S_ERROR("Can't load %s" % certLoc)
    retVal = chain.loadKeyFromFile(keyLoc, password=params.userPasswd)
    if not retVal["OK"]:
        gLogger.warn(retVal["Message"])
        if "bad decrypt" in retVal["Message"] or "bad pass phrase" in retVal["Message"]:
            return S_ERROR("Bad passphrase")
        return S_ERROR("Can't load %s" % keyLoc)

    if params.checkWithCS:
        retVal = chain.generateProxyToFile(
            proxyLoc, params.proxyLifeTime, strength=params.proxyStrength, limited=params.limitedProxy, rfc=params.rfc
        )

        gLogger.info("Contacting CS...")
        retVal = Script.enableCS()
        if not retVal["OK"]:
            gLogger.warn(retVal["Message"])
            if "Unauthorized query" in retVal["Message"]:
                # add hint for users
                return S_ERROR(
                    "Can't contact DIRAC CS: %s (User possibly not registered with dirac server) " % retVal["Message"]
                )
            return S_ERROR("Can't contact DIRAC CS: %s" % retVal["Message"])
        userDN = chain.getCertInChain(-1)["Value"].getSubjectDN()["Value"]

        if not params.diracGroup:
            result = Registry.findDefaultGroupForDN(userDN)
            if not result["OK"]:
                gLogger.warn("Could not get a default group for DN %s: %s" % (userDN, result["Message"]))
            else:
                params.diracGroup = result["Value"]
                gLogger.info("Default discovered group is %s" % params.diracGroup)
        gLogger.info("Checking DN %s" % userDN)
        retVal = Registry.getUsernameForDN(userDN)
        if not retVal["OK"]:
            gLogger.warn(retVal["Message"])
            return S_ERROR("DN %s is not registered" % userDN)
        username = retVal["Value"]
        gLogger.info("Username is %s" % username)
        retVal = Registry.getGroupsForUser(username)
        if not retVal["OK"]:
            gLogger.warn(retVal["Message"])
            return S_ERROR("User %s has no groups defined" % username)
        groups = retVal["Value"]
        if params.diracGroup not in groups:
            return S_ERROR("Requested group %s is not valid for DN %s" % (params.diracGroup, userDN))
        gLogger.info("Creating proxy for %s@%s (%s)" % (username, params.diracGroup, userDN))
    if params.summary:
        h = int(params.proxyLifeTime / 3600)
        m = int(params.proxyLifeTime / 60) - h * 60
        gLogger.notice("Proxy lifetime will be %02d:%02d" % (h, m))
        gLogger.notice("User cert is %s" % certLoc)
        gLogger.notice("User key  is %s" % keyLoc)
        gLogger.notice("Proxy will be written to %s" % proxyLoc)
        if params.diracGroup:
            gLogger.notice("DIRAC Group will be set to %s" % params.diracGroup)
        else:
            gLogger.notice("No DIRAC Group will be set")
        gLogger.notice("Proxy strength will be %s" % params.proxyStrength)
        if params.limitedProxy:
            gLogger.notice("Proxy will be limited")
    retVal = chain.generateProxyToFile(
        proxyLoc,
        params.proxyLifeTime,
        params.diracGroup,
        strength=params.proxyStrength,
        limited=params.limitedProxy,
        rfc=params.rfc,
    )
    if not retVal["OK"]:
        gLogger.warn(retVal["Message"])
        return S_ERROR("Couldn't generate proxy: %s" % retVal["Message"])
    return S_OK(proxyLoc)



"""A generic client for creating and managing transformations.

See the information about transformation parameters below.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import six
import json

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities.JEncode import encode
from DIRAC.Core.Utilities.PromptUser import promptUser
from DIRAC.Core.Base.API import API
from DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody import BaseBody
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.RequestManagementSystem.Client.Operation import Operation

COMPONENT_NAME = "Transformation"

__RCSID__ = "$Id$"


class Transformation(API):

    #############################################################################
    def __init__(self, transID=0, transClient=None):
        """c'tor"""
        super(Transformation, self).__init__()

        self.paramTypes = {
            "TransformationID": six.integer_types,
            "TransformationName": six.string_types,
            "Status": six.string_types,
            "Description": six.string_types,
            "LongDescription": six.string_types,
            "Type": six.string_types,
            "Plugin": six.string_types,
            "AgentType": six.string_types,
            "FileMask": six.string_types,
            "TransformationGroup": six.string_types,
            "GroupSize": six.integer_types + (float,),
            "InheritedFrom": six.integer_types,
            "Body": six.string_types,
            "MaxNumberOfTasks": six.integer_types,
            "EventsPerTask": six.integer_types,
        }
        self.paramValues = {
            "TransformationID": 0,
            "TransformationName": "",
            "Status": "New",
            "Description": "",
            "LongDescription": "",
            "Type": "",
            "Plugin": "Standard",
            "AgentType": "Manual",
            "FileMask": "",
            "TransformationGroup": "General",
            "GroupSize": 1,
            "InheritedFrom": 0,
            "Body": "",
            "MaxNumberOfTasks": 0,
            "EventsPerTask": 0,
        }

        # the metaquery parameters are neither part of the transformation parameters nor the additional parameters, so
        # special treatment is necessary
        self.inputMetaQuery = None
        self.outputMetaQuery = None

        self.ops = Operations()
        self.supportedPlugins = self.ops.getValue(
            "Transformations/AllowedPlugins", ["Broadcast", "Standard", "BySize", "ByShare"]
        )
        if not transClient:
            self.transClient = TransformationClient()
        else:
            self.transClient = transClient
        self.serverURL = self.transClient.getServer()
        self.exists = False
        if transID:
            self.paramValues["TransformationID"] = transID
            res = self.getTransformation()
            if res["OK"]:
                self.exists = True
            elif res["Message"] == "Transformation does not exist":
                raise AttributeError("TransformationID %d does not exist" % transID)
            else:
                self.paramValues["TransformationID"] = 0
                gLogger.fatal(
                    "Failed to get transformation from database", "%s @ %s" % (transID, self.transClient.serverURL)
                )

    def getServer(self):
        return self.serverURL

    def reset(self, transID=0):
        self.__init__(transID)
        self.transClient.setServer(self.serverURL)
        return S_OK()

    def setTargetSE(self, seList):
        return self.__setSE("TargetSE", seList)

    def setSourceSE(self, seList):
        return self.__setSE("SourceSE", seList)

    def setBody(self, body):
        """check that the body is a string, or using the proper syntax for multiple operations,
        or is a BodyPlugin object

        :param body: transformation body, for example

          .. code :: python

            body = [ ( "ReplicateAndRegister", { "SourceSE":"FOO-SRM", "TargetSE":"BAR-SRM" }),
                     ( "RemoveReplica", { "TargetSE":"FOO-SRM" } ),
                   ]

        :type body: string or list of tuples (or lists) of string and dictionaries or a Body plugin (:py:class:`DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody.BaseBody`)
        :raises TypeError: If the structure is not as expected
        :raises ValueError: If unknown attribute for the :class:`~DIRAC.RequestManagementSystem.Client.Operation.Operation`
                            is used
        :returns: S_OK, S_ERROR
        """
        self.item_called = "Body"

        # Simple single operation body case
        if isinstance(body, six.string_types):
            return self.__setParam(body)

        # BodyPlugin case
        elif isinstance(body, BaseBody):
            return self.__setParam(encode(body))

        if not isinstance(body, (list, tuple)):
            raise TypeError("Expected list or string, but %r is %s" % (body, type(body)))

        # MultiOperation body case
        for tup in body:
            if not isinstance(tup, (tuple, list)):
                raise TypeError("Expected tuple or list, but %r is %s" % (tup, type(tup)))
            if len(tup) != 2:
                raise TypeError("Expected 2-tuple, but %r is length %d" % (tup, len(tup)))
            if not isinstance(tup[0], six.string_types):
                raise TypeError("Expected string, but first entry in tuple %r is %s" % (tup, type(tup[0])))
            if not isinstance(tup[1], dict):
                raise TypeError("Expected dictionary, but second entry in tuple %r is %s" % (tup, type(tup[0])))
            for par, val in tup[1].items():
                if not isinstance(par, six.string_types):
                    raise TypeError("Expected string, but key in dictionary %r is %s" % (par, type(par)))
                if par not in Operation.ATTRIBUTE_NAMES:
                    raise ValueError("Unknown attribute for Operation: %s" % par)
                if not isinstance(val, six.string_types + six.integer_types + (float, list, tuple, dict)):
                    raise TypeError("Cannot encode %r, in json" % (val))
        return self.__setParam(json.dumps(body))

    def setInputMetaQuery(self, query):
        """Set the input meta query.

        :param dict query: dictionary to use for input meta query
        """
        self.inputMetaQuery = query
        return S_OK()

    def setOutputMetaQuery(self, query):
        """Set the output meta query.

        :param dict query: dictionary to use for output meta query
        """
        self.outputMetaQuery = query
        return S_OK()

    def __setSE(self, seParam, seList):
        if isinstance(seList, six.string_types):
            try:
                seList = eval(seList)
            except Exception:
                seList = seList.split(",")
        elif isinstance(seList, (list, dict, tuple)):
            seList = list(seList)
        else:
            return S_ERROR("Bad argument type")
        res = self.__checkSEs(seList)
        if not res["OK"]:
            return res
        self.item_called = seParam
        return self.__setParam(seList)

    def __getattr__(self, name):
        if name.find("get") == 0:
            item = name[3:]
            self.item_called = item
            return self.__getParam
        if name.find("set") == 0:
            item = name[3:]
            self.item_called = item
            return self.__setParam
        raise AttributeError(name)

    def __getParam(self):
        if self.item_called == "Available":
            return S_OK(list(self.paramTypes))
        if self.item_called == "Parameters":
            return S_OK(self.paramValues)
        if self.item_called in self.paramValues:
            return S_OK(self.paramValues[self.item_called])
        raise AttributeError("Unknown parameter for transformation: %s" % self.item_called)

    def __setParam(self, value):
        change = False
        if self.item_called in self.paramTypes:
            if self.paramValues[self.item_called] != value:
                if isinstance(value, self.paramTypes[self.item_called]):
                    change = True
                else:
                    raise TypeError(
                        "%s %s %s expected one of %s"
                        % (self.item_called, value, type(value), self.paramTypes[self.item_called])
                    )
        else:
            if self.item_called not in self.paramValues:
                change = True
            else:
                if self.paramValues[self.item_called] != value:
                    change = True
        if not change:
            gLogger.verbose("No change of parameter %s required" % self.item_called)
        else:
            gLogger.verbose("Parameter %s to be changed" % self.item_called)
            transID = self.paramValues["TransformationID"]
            if self.exists and transID:
                res = self.transClient.setTransformationParameter(transID, self.item_called, value)
                if not res["OK"]:
                    return res
            self.paramValues[self.item_called] = value
        return S_OK()

    def getTransformation(self, printOutput=False):
        transID = self.paramValues["TransformationID"]
        if not transID:
            gLogger.fatal("No TransformationID known")
            return S_ERROR()
        res = self.transClient.getTransformation(transID, extraParams=True)
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res
        transParams = res["Value"]
        for paramName, paramValue in transParams.items():
            setter = None
            setterName = "set%s" % paramName
            if hasattr(self, setterName) and callable(getattr(self, setterName)):
                setter = getattr(self, setterName)
            if not setterName:
                gLogger.error("Unable to invoke setter %s, it isn't a member function" % setterName)
                continue
            setter(paramValue)
        if printOutput:
            gLogger.info("No printing available yet")
        return S_OK(transParams)

    def getTransformationLogging(self, printOutput=False):
        transID = self.paramValues["TransformationID"]
        if not transID:
            gLogger.fatal("No TransformationID known")
            return S_ERROR()
        res = self.transClient.getTransformationLogging(transID)
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res
        loggingList = res["Value"]
        if printOutput:
            self._printFormattedDictList(
                loggingList, ["Message", "MessageDate", "AuthorDN"], "MessageDate", "MessageDate"
            )
        return S_OK(loggingList)

    def extendTransformation(self, nTasks, printOutput=False):
        return self.__executeOperation("extendTransformation", nTasks, printOutput=printOutput)

    def cleanTransformation(self, printOutput=False):
        res = self.__executeOperation("cleanTransformation", printOutput=printOutput)
        if res["OK"]:
            self.paramValues["Status"] = "Cleaned"
        return res

    def deleteTransformation(self, printOutput=False):
        res = self.__executeOperation("deleteTransformation", printOutput=printOutput)
        if res["OK"]:
            self.reset()
        return res

    def addFilesToTransformation(self, lfns, printOutput=False):
        return self.__executeOperation("addFilesToTransformation", lfns, printOutput=printOutput)

    def setFileStatusForTransformation(self, status, lfns, printOutput=False):
        return self.__executeOperation("setFileStatusForTransformation", status, lfns, printOutput=printOutput)

    def getTransformationTaskStats(self, printOutput=False):
        return self.__executeOperation("getTransformationTaskStats", printOutput=printOutput)

    def getTransformationStats(self, printOutput=False):
        return self.__executeOperation("getTransformationStats", printOutput=printOutput)

    def deleteTasks(self, taskMin, taskMax, printOutput=False):
        return self.__executeOperation("deleteTasks", taskMin, taskMax, printOutput=printOutput)

    def addTaskForTransformation(self, lfns=[], se="Unknown", printOutput=False):
        return self.__executeOperation("addTaskForTransformation", lfns, se, printOutput=printOutput)

    def setTaskStatus(self, taskID, status, printOutput=False):
        return self.__executeOperation("setTaskStatus", taskID, status, printOutput=printOutput)

    def __executeOperation(self, operation, *parms, **kwds):
        transID = self.paramValues["TransformationID"]
        if not transID:
            gLogger.fatal("No TransformationID known")
            return S_ERROR()
        printOutput = kwds.pop("printOutput")
        fcn = None
        if hasattr(self.transClient, operation) and callable(getattr(self.transClient, operation)):
            fcn = getattr(self.transClient, operation)
        if not fcn:
            return S_ERROR("Unable to invoke %s, it isn't a member funtion of TransformationClient")
        res = fcn(transID, *parms, **kwds)
        if printOutput:
            self._prettyPrint(res)
        return res

    def getTransformationFiles(
        self,
        fileStatus=[],
        lfns=[],
        outputFields=[
            "FileID",
            "LFN",
            "Status",
            "TaskID",
            "TargetSE",
            "UsedSE",
            "ErrorCount",
            "InsertedTime",
            "LastUpdate",
        ],
        orderBy="FileID",
        printOutput=False,
    ):
        condDict = {"TransformationID": self.paramValues["TransformationID"]}
        if fileStatus:
            condDict["Status"] = fileStatus
        if lfns:
            condDict["LFN"] = lfns
        res = self.transClient.getTransformationFiles(condDict=condDict)
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res
        if printOutput:
            if not outputFields:
                gLogger.info("Available fields are: %s" % res["ParameterNames"].join(" "))
            elif not res["Value"]:
                gLogger.info("No tasks found for selection")
            else:
                self._printFormattedDictList(res["Value"], outputFields, "FileID", orderBy)
        return res

    def getTransformationTasks(
        self,
        taskStatus=[],
        taskIDs=[],
        outputFields=[
            "TransformationID",
            "TaskID",
            "ExternalStatus",
            "ExternalID",
            "TargetSE",
            "CreationTime",
            "LastUpdateTime",
        ],
        orderBy="TaskID",
        printOutput=False,
    ):
        condDict = {"TransformationID": self.paramValues["TransformationID"]}
        if taskStatus:
            condDict["ExternalStatus"] = taskStatus
        if taskIDs:
            condDict["TaskID"] = taskIDs
        res = self.transClient.getTransformationTasks(condDict=condDict)
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res
        if printOutput:
            if not outputFields:
                gLogger.info("Available fields are: %s" % res["ParameterNames"].join(" "))
            elif not res["Value"]:
                gLogger.info("No tasks found for selection")
            else:
                self._printFormattedDictList(res["Value"], outputFields, "TaskID", orderBy)
        return res

    #############################################################################
    def getTransformations(
        self,
        transID=[],
        transStatus=[],
        outputFields=["TransformationID", "Status", "AgentType", "TransformationName", "CreationDate"],
        orderBy="TransformationID",
        printOutput=False,
    ):
        condDict = {}
        if transID:
            condDict["TransformationID"] = transID
        if transStatus:
            condDict["Status"] = transStatus
        res = self.transClient.getTransformations(condDict=condDict)
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res
        if printOutput:
            if not outputFields:
                gLogger.info("Available fields are: %s" % res["ParameterNames"].join(" "))
            elif not res["Value"]:
                gLogger.info("No tasks found for selection")
            else:
                self._printFormattedDictList(res["Value"], outputFields, "TransformationID", orderBy)
        return res

    #############################################################################
    def getAuthorDNfromProxy(self):
        """gets the AuthorDN and username of the transformation from the uploaded proxy"""
        username = ""
        author = ""
        res = getProxyInfo()
        if res["OK"]:
            author = res["Value"]["identity"]
            username = res["Value"]["username"]
        else:
            gLogger.error("Unable to get uploaded proxy Info %s " % res["Message"])
            return S_ERROR(res["Message"])

        res = {"username": username, "authorDN": author}
        return S_OK(res)

    #############################################################################
    def getTransformationsByUser(
        self,
        authorDN="",
        userName="",
        transID=[],
        transStatus=[],
        outputFields=["TransformationID", "Status", "AgentType", "TransformationName", "CreationDate", "AuthorDN"],
        orderBy="TransformationID",
        printOutput=False,
    ):
        condDict = {}
        if authorDN == "":
            res = self.getAuthorDNfromProxy()
            if not res["OK"]:
                gLogger.error(res["Message"])
                return S_ERROR(res["Message"])
            else:
                foundUserName = res["Value"]["username"]
                foundAuthor = res["Value"]["authorDN"]
                # If the username whom created the uploaded proxy is different than the provided username report error and exit
                if not (userName == "" or userName == foundUserName):
                    gLogger.error(
                        "Couldn't resolve the authorDN for user '%s' from the uploaded proxy (proxy created by '%s')"
                        % (userName, foundUserName)
                    )
                    return S_ERROR(
                        "Couldn't resolve the authorDN for user '%s' from the uploaded proxy (proxy created by '%s')"
                        % (userName, foundUserName)
                    )

                userName = foundUserName
                authorDN = foundAuthor
                gLogger.info(
                    "Will list transformations created by user '%s' with status '%s'"
                    % (userName, ", ".join(transStatus))
                )
        else:
            gLogger.info(
                "Will list transformations created by '%s' with status '%s'" % (authorDN, ", ".join(transStatus))
            )

        condDict["AuthorDN"] = authorDN
        if transID:
            condDict["TransformationID"] = transID
        if transStatus:
            condDict["Status"] = transStatus
        res = self.transClient.getTransformations(condDict=condDict)
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res

        if printOutput:
            if not outputFields:
                gLogger.info("Available fields are: %s" % res["ParameterNames"].join(" "))
            elif not res["Value"]:
                gLogger.info("No tasks found for selection")
            else:
                self._printFormattedDictList(res["Value"], outputFields, "TransformationID", orderBy)
        return res

    #############################################################################
    def getSummaryTransformations(self, transID=[]):
        """Show the summary for a list of Transformations

        Fields starting with 'F' ('J')  refers to files (jobs).
        Proc. stand for processed.
        """
        condDict = {"TransformationID": transID}
        orderby = []
        start = 0
        maxitems = len(transID)
        paramShowNames = [
            "TransformationID",
            "Type",
            "Status",
            "Files_Total",
            "Files_PercentProcessed",
            "Files_Processed",
            "Files_Unused",
            "Jobs_TotalCreated",
            "Jobs_Waiting",
            "Jobs_Running",
            "Jobs_Done",
            "Jobs_Failed",
            "Jobs_Stalled",
        ]
        # Below, the header used for each field in the printing: short to fit in one line
        paramShowNamesShort = [
            "TransID",
            "Type",
            "Status",
            "F_Total",
            "F_Proc.(%)",
            "F_Proc.",
            "F_Unused",
            "J_Created",
            "J_Wait",
            "J_Run",
            "J_Done",
            "J_Fail",
            "J_Stalled",
        ]
        dictList = []

        result = self.transClient.getTransformationSummaryWeb(condDict, orderby, start, maxitems)
        if not result["OK"]:
            self._prettyPrint(result)
            return result

        if result["Value"]["TotalRecords"] > 0:
            try:
                paramNames = result["Value"]["ParameterNames"]
                for paramValues in result["Value"]["Records"]:
                    paramShowValues = map(lambda pname: paramValues[paramNames.index(pname)], paramShowNames)
                    showDict = dict(zip(paramShowNamesShort, paramShowValues))
                    dictList.append(showDict)

            except Exception as x:
                print("Exception %s " % str(x))

        if not len(dictList) > 0:
            gLogger.error("No found transformations satisfying input condition")
            return S_ERROR("No found transformations satisfying input condition")
        else:
            print(
                self._printFormattedDictList(
                    dictList, paramShowNamesShort, paramShowNamesShort[0], paramShowNamesShort[0]
                )
            )

        return S_OK(dictList)

    #############################################################################
    def addTransformation(self, addFiles=True, printOutput=False):
        """Add transformation to the transformation system.

        Sets all parameters currently assigned to the transformation.

        :param bool addFiles: if True, immediately perform input data query
        :param bool printOutput: if True, print information about transformation
        """
        res = self._checkCreation()
        if not res["OK"]:
            return self._errorReport(res, "Failed transformation sanity check")
        if printOutput:
            gLogger.info("Will attempt to create transformation with the following parameters")
            self._prettyPrint(self.paramValues)

        res = self.transClient.addTransformation(
            self.paramValues["TransformationName"],
            self.paramValues["Description"],
            self.paramValues["LongDescription"],
            self.paramValues["Type"],
            self.paramValues["Plugin"],
            self.paramValues["AgentType"],
            self.paramValues["FileMask"],
            transformationGroup=self.paramValues["TransformationGroup"],
            groupSize=self.paramValues["GroupSize"],
            inheritedFrom=self.paramValues["InheritedFrom"],
            body=self.paramValues["Body"],
            maxTasks=self.paramValues["MaxNumberOfTasks"],
            eventsPerTask=self.paramValues["EventsPerTask"],
            addFiles=addFiles,
            inputMetaQuery=self.inputMetaQuery,
            outputMetaQuery=self.outputMetaQuery,
        )
        if not res["OK"]:
            if printOutput:
                self._prettyPrint(res)
            return res
        transID = res["Value"]
        self.exists = True
        self.setTransformationID(transID)
        gLogger.notice("Created transformation %d" % transID)
        for paramName, paramValue in self.paramValues.items():
            if paramName not in self.paramTypes:
                res = self.transClient.setTransformationParameter(transID, paramName, paramValue)
                if not res["OK"]:
                    gLogger.error("Failed to add parameter", "%s %s" % (paramName, res["Message"]))
                    gLogger.notice("To add this parameter later please execute the following.")
                    gLogger.notice("oTransformation = Transformation(%d)" % transID)
                    gLogger.notice("oTransformation.set%s(...)" % paramName)
        return S_OK(transID)

    def _checkCreation(self):
        """Few checks"""
        if self.paramValues["TransformationID"]:
            gLogger.info("You are currently working with an active transformation definition.")
            gLogger.info("If you wish to create a new transformation reset the TransformationID.")
            gLogger.info("oTransformation.reset()")
            return S_ERROR()

        requiredParameters = ["TransformationName", "Description", "LongDescription", "Type"]
        for parameter in requiredParameters:
            if not self.paramValues[parameter]:
                gLogger.info("%s is not defined for this transformation. This is required..." % parameter)
                self.paramValues[parameter] = six.moves.input("Please enter the value of " + parameter + " ")

        plugin = self.paramValues["Plugin"]
        if plugin:
            if plugin not in self.supportedPlugins:
                gLogger.info("The selected Plugin (%s) is not known to the transformation agent." % plugin)
                res = self.__promptForParameter("Plugin", choices=self.supportedPlugins, default="Standard")
                if not res["OK"]:
                    return res
                self.paramValues["Plugin"] = res["Value"]

        plugin = self.paramValues["Plugin"]

        return S_OK()

    def _checkBySizePlugin(self):
        return self._checkStandardPlugin()

    def _checkBySharePlugin(self):
        return self._checkStandardPlugin()

    def _checkStandardPlugin(self):
        groupSize = self.paramValues["GroupSize"]
        if groupSize <= 0:
            gLogger.info("The GroupSize was found to be less than zero. It has been set to 1.")
            res = self.setGroupSize(1)
            if not res["OK"]:
                return res
        return S_OK()

    def _checkBroadcastPlugin(self):
        gLogger.info(
            "The Broadcast plugin requires the following parameters be set: %s" % (", ".join(["SourceSE", "TargetSE"]))
        )
        requiredParams = ["SourceSE", "TargetSE"]
        for requiredParam in requiredParams:
            if not self.paramValues.get(requiredParam):
                paramValue = six.moves.input("Please enter " + requiredParam + " ")
                setter = None
                setterName = "set%s" % requiredParam
                if hasattr(self, setterName) and callable(getattr(self, setterName)):
                    setter = getattr(self, setterName)
                if not setter:
                    return S_ERROR("Unable to invoke %s, this function hasn't been implemented." % setterName)
                ses = paramValue.replace(",", " ").split()
                res = setter(ses)
                if not res["OK"]:
                    return res
        return S_OK()

    def __checkSEs(self, seList):
        res = gConfig.getSections("/Resources/StorageElements")
        if not res["OK"]:
            return self._errorReport(res, "Failed to get possible StorageElements")
        missing = set(seList) - set(res["Value"])
        if missing:
            for se in missing:
                gLogger.error("StorageElement %s is not known" % se)
            return S_ERROR("%d StorageElements not known" % len(missing))
        return S_OK()

    def __promptForParameter(self, parameter, choices=[], default="", insert=True):
        res = promptUser("Please enter %s" % parameter, choices=choices, default=default)
        if not res["OK"]:
            return self._errorReport(res)
        gLogger.notice("%s will be set to '%s'" % (parameter, res["Value"]))
        paramValue = res["Value"]
        if insert:
            setter = None
            setterName = "set%s" % parameter
            if hasattr(self, setterName) and callable(getattr(self, setterName)):
                setter = getattr(self, setterName)
            if not setter:
                return S_ERROR("Unable to invoke %s, it isn't a member function of Transformation!")
            res = setter(paramValue)
            if not res["OK"]:
                return res
        return S_OK(paramValue)
